# -*- coding: utf-8 -*-


from libs import bflb_utils

keyslot0 = 28
keyslot1 = keyslot0 + 16
keyslot2 = keyslot1 + 16
keyslot3 = keyslot2 + 16
keyslot4 = keyslot3 + 16
keyslot5 = keyslot4 + 16
keyslot6 = keyslot5 + 16

wr_lock_key_slot_4_l = 13
wr_lock_key_slot_5_l = 14
wr_lock_boot_mode = 15
wr_lock_dbg_pwd = 16
wr_lock_sw_usage_0 = 17
wr_lock_wifi_mac = 18
wr_lock_key_slot_0 = 19
wr_lock_key_slot_1 = 20
wr_lock_key_slot_2 = 21
wr_lock_key_slot_3 = 22
wr_lock_key_slot_4_h = 23
wr_lock_key_slot_5_h = 24
rd_lock_dbg_pwd = 25
rd_lock_key_slot_0 = 26
rd_lock_key_slot_1 = 27
rd_lock_key_slot_2 = 28
rd_lock_key_slot_3 = 29
rd_lock_key_slot_4 = 30
rd_lock_key_slot_5 = 31


def efuse_data_create(name, value):
    efuse_data = bytearray(128)
    efuse_data_mask = bytearray(128)
    mask_4bytes = bytearray.fromhex("FFFFFFFF")
    rw_lock = 0
    if name == "flash_aes_key":
        # encrypt type
        if len(value[0]) == 32:
            efuse_data[0] |= 0x01
        if len(value[0]) == 48:
            efuse_data[0] |= 0x02
        if len(value[0]) == 64:
            efuse_data[0] |= 0x03
        # encrypt flag
        efuse_data[0] |= 0x80
        efuse_data_mask[0] |= 0xff
        # encrypt key
        keydata = bytearray.fromhex(value[0])
        efuse_data[keyslot2:keyslot4] = keydata + bytearray(32 - len(keydata))
        efuse_data_mask[keyslot2:keyslot4] = mask_4bytes * 8
        # encrypt key read/write lock
        rw_lock |= (1 << wr_lock_key_slot_2)
        rw_lock |= (1 << wr_lock_key_slot_3)
        rw_lock |= (1 << rd_lock_key_slot_2)
        rw_lock |= (1 << rd_lock_key_slot_3)
        efuse_data[124:128] = bflb_utils.int_to_4bytearray_l(rw_lock)
        efuse_data_mask[124:128] = bflb_utils.int_to_4bytearray_l(rw_lock)
    return efuse_data, efuse_data_mask



# -*- coding:utf-8 -*-
#  Copyright (C) 2021- BOUFFALO LAB (NANJING) CO., LTD.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.


import time
import binascii
import serial
import threading
from queue import Queue
from queue import Empty

from libs import bflb_utils


class BflbUartPort(object):

    def __init__(self):
        self._baudrate = 115200
        self._ser = None
        self._shakehand_flag = False
        self._chiptype = "bl602"

    def if_init(self, device, rate, chiptype="bl602"):
        if self._ser is None:
            self._baudrate = rate
            self._ser = serial.Serial(device,
                                      rate,
                                      timeout=2.0,
                                      xonxoff=False,
                                      rtscts=False,
                                      write_timeout=None,
                                      dsrdtr=False)
        else:
            self._ser.baudrate = rate
            self._baudrate = rate
        self._602a0_dln_fix = False
        self._chiptype = chiptype

    def if_set_rx_timeout(self, val):
        self._ser.timeout = val

    def if_set_dtr(self, val):
        self._ser.setDTR(val)

    def if_set_rts(self, val):
        self._ser.setRTS(val)

    def if_get_rate(self):
        return self._baudrate

    def if_write(self, data_send):
        # bflb_utils.printf("sending ", binascii.hexlify(data_send))
        self._ser.write(data_send)

    def if_raw_read(self):
        return self._ser.read(self._ser.in_waiting or 1)

    # this function return bytearray or bytes type
    def if_read(self, data_len):
        data = bytearray(0)
        received = 0
        while received < data_len:
            tmp = self._ser.read(data_len - received)
            if len(tmp) == 0:
                break
            data += tmp
            received += len(tmp)

        if len(data) != data_len:
            return 0, data
        return 1, data

    def _if_get_sync_bytes(self, length):
        data = bytearray(length)
        i = 0
        while i < length:
            data[i] = 0x55
            i += 1
        return data

    def if_set_602a0_download_fix(self, val):
        self._602a0_dln_fix = val

    def if_send_55(self):
        while True:
            if self._shakehand_flag is True:
                break
            if self._chiptype == "bl702":
                self._ser.write(self._if_get_sync_bytes(int(0.003 * self._baudrate / 10)))
            else:
                self._ser.write(self._if_get_sync_bytes(int(0.006 * self._baudrate / 10)))

    # this function return str type
    def if_shakehand(self,
                     do_reset=False,
                     reset_hold_time=100,
                     shake_hand_delay=100,
                     reset_revert=True,
                     cutoff_time=0,
                     shake_hand_retry=2,
                     boot2_load=0):
        timeout = self._ser.timeout
        if boot2_load > 0:
            wait_timeout = boot2_load
            self._ser.timeout = 0.1
            self._shakehand_flag = False
            fl_thrx = None
            fl_thrx = threading.Thread(target=self.if_send_55)
            fl_thrx.setDaemon(True)
            fl_thrx.start()
            time_stamp = time.time()
            while time.time() - time_stamp < wait_timeout:
                success, ack = self.if_read(16)
                if ack.find(b'\x4F') != -1 or ack.find(b'\x4B') != -1:
                    self._ser.timeout = timeout
                    time.sleep(0.03)
                    self._shakehand_flag = True
                    return "OK"
            self._shakehand_flag = True
            self._ser.timeout = timeout
            return "FL"
        while shake_hand_retry > 0:
            # cut of tx rx power and rst
            if cutoff_time != 0:
                cutoff_revert = False
                if cutoff_time > 1000:
                    cutoff_revert = True
                    cutoff_time = cutoff_time - 1000
                # MP_TOOL_V3 generate rising pulse to make D trigger output low
                # reset low
                self._ser.setRTS(1)
                # RC delay is 100ms
                time.sleep(0.2)
                # reset high
                self._ser.setRTS(0)
                time.sleep(0.05)
                # do power off
                # reset low
                self._ser.setRTS(1)
                if cutoff_revert:
                    # dtr high, power off
                    self._ser.setDTR(0)
                else:
                    # dtr low, power off
                    self._ser.setDTR(1)
                bflb_utils.printf("tx rx and power off, press the machine!")
                bflb_utils.printf("cutoff time is ", cutoff_time / 1000.0)
                time.sleep(cutoff_time / 1000.0)
                if cutoff_revert:
                    # dtr low, power on
                    self._ser.setDTR(1)
                else:
                    # dtr high, power on
                    self._ser.setDTR(0)
                bflb_utils.printf("power on tx and rx ")
                time.sleep(0.1)
            if do_reset is True:
                # MP_TOOL_V3 reset high to make boot pin high
                self._ser.setRTS(0)
                time.sleep(0.2)
                if reset_revert:
                    # reset low for reset revert to make boot pin high when cpu rset
                    self._ser.setRTS(1)
                    time.sleep(0.001)
                reset_cnt = 2
                if reset_hold_time > 1000:
                    reset_cnt = int(reset_hold_time // 1000)
                    reset_hold_time = reset_hold_time % 1000
                while reset_cnt > 0:
                    if reset_revert:
                        # reset high
                        self._ser.setRTS(0)
                    else:
                        # reset low
                        self._ser.setRTS(1)
                    # Boot high
                    # self._ser.setDTR(0)
                    time.sleep(reset_hold_time / 1000.0)
                    if reset_revert:
                        # reset low
                        self._ser.setRTS(1)
                    else:
                        # reset high
                        self._ser.setRTS(0)
                    if shake_hand_delay > 0:
                        time.sleep(shake_hand_delay / 1000.0)
                    else:
                        time.sleep(5 / 1000.0)

                    # do reset agian to make sure boot pin is high
                    if reset_revert:
                        # reset high
                        self._ser.setRTS(0)
                    else:
                        # reset low
                        self._ser.setRTS(1)
                    # Boot high
                    # self._ser.setDTR(0)
                    time.sleep(reset_hold_time / 1000.0)
                    if reset_revert:
                        # reset low
                        self._ser.setRTS(1)
                    else:
                        # reset high
                        self._ser.setRTS(0)
                    if shake_hand_delay > 0:
                        time.sleep(shake_hand_delay / 1000.0)
                    else:
                        time.sleep(5 / 1000.0)
                    reset_cnt -= 1
                bflb_utils.printf("reset cnt: " + str(reset_cnt) + ", reset hold: " +
                                  str(reset_hold_time / 1000.0) + ", shake hand delay: " +
                                  str(shake_hand_delay / 1000.0))
            # clean buffer before start
            bflb_utils.printf("clean buf")
            self._ser.timeout = 0.1
            success, ack = self.if_read(1000)
            # change tiemout value when shake hand
            if self._602a0_dln_fix:
                self._ser.timeout = 0.5
            else:
                self._ser.timeout = 0.1
            bflb_utils.printf("send sync")
            # send keep 6ms ,N*10/baudrate=0.01
            if self._chiptype == "bl702":
                self._ser.write(self._if_get_sync_bytes(int(0.003 * self._baudrate / 10)))
            else:
                self._ser.write(self._if_get_sync_bytes(int(0.006 * self._baudrate / 10)))
            if self._602a0_dln_fix:
                time.sleep(4)
            success, ack = self.if_read(16)
            bflb_utils.printf("ack is ", binascii.hexlify(ack))
            if ack.find(b'\x4F') != -1 or ack.find(b'\x4B') != -1:
                self._ser.timeout = timeout
                if self._602a0_dln_fix:
                    self._ser.write(bytearray(2))
                time.sleep(0.03)
                return "OK"
            if len(ack) != 0:
                # peer is alive, but shake hand it's not expected, do again
                bflb_utils.printf("reshake")
                if do_reset is False:
                    bflb_utils.printf("sleep")
                    #tmp=0
                    #while tmp<60:
                    #    time.sleep(2)
                    #    bflb_utils.printf("Pls check eflash loader log")
                    #    tmp+=1
                    time.sleep(3)
            else:
                bflb_utils.printf("retry")
            shake_hand_retry -= 1
        self._ser.timeout = timeout
        return "FL"

    def if_close(self):
        if self._ser is not None:
            self._ser.close()
            self._ser = None

    # this function return str type
    def if_deal_ack(self, dmy_data=True):
        success, ack = self.if_read(2)
        if success == 0:
            bflb_utils.printf("ack is ", str(binascii.hexlify(ack)))
            return ack.decode("utf-8")
        if ack.find(b'\x4F') != -1 or ack.find(b'\x4B') != -1:
            #if dmy_data:
            #    success,ack=self.if_read(14)
            #    if success==0:
            #        return "FL"
            return "OK"
        success, err_code = self.if_read(2)
        if success == 0:
            bflb_utils.printf("err code is ", str(binascii.hexlify(err_code)))
            return "FL"
        err_code_str = str(binascii.hexlify(err_code[1:2] + err_code[0:1]).decode('utf-8'))
        ack = "FL"
        try:
            ret = ack + err_code_str + "(" + bflb_utils.get_bflb_error_code[err_code_str] + ")"
        except Exception:
            ret = ack + err_code_str + " unknown"
        bflb_utils.printf(ret)
        return ret

    # this function return bytearray or bytes type
    def if_deal_response(self):
        ack = self.if_deal_ack()
        if ack == "OK":
            success, len_bytes = self.if_read(2)
            if success == 0:
                bflb_utils.printf("Get length error")
                bflb_utils.printf("len error is ", binascii.hexlify(len_bytes))
                return "Get length error", len_bytes
            tmp = bflb_utils.bytearray_reverse(len_bytes)
            data_len = bflb_utils.bytearray_to_int(tmp)
            success, data_bytes = self.if_read(data_len)
            if success == 0 or len(data_bytes) != data_len:
                bflb_utils.printf("Read data error,maybe not get excepted length")
                # bflb_utils.printf("ack is ", str(binascii.hexlify(data_bytes)))
                return "Read data error,maybe not get excepted length", data_bytes
            # bflb_utils.printf("ack is ", str(binascii.hexlify(data_bytes)))
            return ack, data_bytes
        bflb_utils.printf("Not ack OK")
        bflb_utils.printf(ack)
        return ack, None


class CliInfUart(object):

    def __init__(self, recv_cb_objs=None):
        self._baudrate = 115200
        self._ser = None
        # self._logfile = None
        self._rx_thread = None
        self._rx_thread_running = False
        self._tx_thread = None
        self._tx_thread = None
        self._tx_queue = None
        if recv_cb_objs is not None:
            self._recv_cb_objs = recv_cb_objs
        else:
            self._recv_cb_objs = []
        self._boot = 0
        self.uart = ""

    def add_observer(self, recv_cb_obj):
        if recv_cb_obj is not None:
            self._recv_cb_objs.append(recv_cb_obj)
            return True
        else:
            return False

    def is_open(self):
        if self._ser is None:
            return False
        else:
            if self._ser.isOpen():
                return True
            else:
                return False

    def open(self, dev_com):
        if not dev_com:
            bflb_utils.printf("No Serial Port Found")
            return False
        if self._ser is None:
            try:
                self._ser = serial.Serial(dev_com,
                                          self._baudrate,
                                          timeout=5.0,
                                          xonxoff=False,
                                          rtscts=False,
                                          write_timeout=None,
                                          dsrdtr=False)
                # self._logfile = mfg_lf.open()
                self._ser.setDTR(1)
                self._ser.setRTS(1)
                time.sleep(0.01)
                self._ser.setRTS(0)
                time.sleep(0.01)
                self._ser.setDTR(0)
                self._tx_queue = Queue()
                self._rx_thread = threading.Thread(target=self._read)
                self._rx_thread_running = True
                self._rx_thread.setDaemon(True)
                self._rx_thread.start()
                self._tx_thread = threading.Thread(target=self._write)
                self._tx_thread_running = True
                self._tx_thread.setDaemon(True)
                self._tx_thread.start()
            except serial.SerialException:
                bflb_utils.printf("Open %s Fail" % (dev_com))
                return False
            except TypeError:
                if self._ser is not None:
                    self._ser.close()
                return False
            else:
                bflb_utils.printf("Open %s Success" % (dev_com))
                return True

    def close(self):
        if self._ser is not None:
            self._ser.close()
            # self._logfile.close()
            self._ser = None
            # self._logfile = None
            self._rx_thread_running = False
            self._rx_thread.join()
            self._tx_thread_running = False
            self._tx_thread.join()
            self._tx_queue = None
            self._boot = 0

    def write(self, data_send):
        # data_send type is bytes
        if self._ser is not None:
            # return self._ser.write(data_send)
            self._tx_queue.put(data_send)
        else:
            return 0

    def _write(self):
        while self._tx_thread_running:
            try:
                data = self._tx_queue.get(timeout=1)
            except Empty:
                pass
            else:
                self._ser.write(data)

    def _check_boot_cond(self, data):

        def bl60x_is_active(data):
            if len(data) > 10:
                return True

        def bl60x_bootup_filter(data):
            if data.find('Compile Date Time') == 0:
                return True
            if data.find('rf init') == 0:
                return True
            return False

        if bl60x_bootup_filter(data) is True:
            return False

        # bl60x start up info
        if data.find('[RX] Statitics') == 0:
            self._boot = 1
            return True

        # app open first, but bl60x has active
        if self._boot == 0 and bl60x_is_active(data):
            self._boot = 1
            # TODO query version
            return True
        return False

    def _read(self):
        data = ''

        while self._rx_thread_running:
            try:
                byte_msg = self._ser.read(1)
            except serial.SerialException:
                try:
                    self._ser.close()
                except Exception:
                    pass
                self._ser = None
                self._rx_thread_running = False
                self._tx_thread_running = False
                self._tx_thread.join()
                self._tx_queue = None
            except Exception:
                pass
            else:
                try:
                    str_msg = bytes.decode(byte_msg)
                except Exception:
                    continue
                if str(str_msg) == '\r':
                    if data != '':
                        bflb_utils.printf(data)
                        # check on_boot
                        if self._check_boot_cond(data) is True:
                            if self._recv_cb_objs is not None and len(self._recv_cb_objs) != 0:
                                for cb in self._recv_cb_objs:
                                    time.sleep(1)
                                    cb.on_boot(self.uart)

                        if self._recv_cb_objs is not None and len(self._recv_cb_objs) != 0:
                            for cb in self._recv_cb_objs:
                                cb.obs_handle(data)
                        # self._logfile.wrtiteline(data)
                    data = ''
                else:
                    if str(str_msg) != '\n':
                        # data += str_msg.strip()
                        data += str_msg

    def open_listen(self, dev_com, baudrate):
        if not dev_com:
            bflb_utils.printf("No Serial Port Found")
            return False
        if self._ser is None:
            try:
                self._ser = serial.Serial(dev_com,
                                          baudrate,
                                          timeout=5.0,
                                          xonxoff=False,
                                          rtscts=False,
                                          write_timeout=None,
                                          dsrdtr=False)
                # self._logfile = mfg_lf.open()
                self._ser.setDTR(1)
                self._ser.setRTS(1)
                time.sleep(0.01)
                self._ser.setRTS(0)
                time.sleep(0.01)
                self._ser.setDTR(0)
                self._rx_thread = threading.Thread(target=self._read_listen)
                self._rx_thread_running = True
                self._rx_thread.setDaemon(True)
                self._rx_thread.start()

            except serial.SerialException:
                bflb_utils.printf("Open %s Listen Fail" % (dev_com))
                return False
            except TypeError:
                if self._ser is not None:
                    self._ser.close()
                return False
            else:
                bflb_utils.printf("Open %s Listen Success" % (dev_com))
                return True

    def close_listen(self):
        if self._ser is not None:
            self._ser.close()
            self._ser = None
            self._rx_thread_running = False
            self._rx_thread.join()

    def _read_listen(self):
        self._ser.timeout = 0.1
        while self._rx_thread_running:
            try:
                byte_msg = self._ser.read(50000)
            except serial.SerialException:
                try:
                    self._ser.close()
                except Exception:
                    pass
                self._ser = None
                self._rx_thread_running = False
            except Exception:
                pass
            else:
                try:
                    if byte_msg:
                        new_mess = str(bytes.decode(byte_msg))
                        bflb_utils.printf(new_mess)
                        byte_msg = ''
                except Exception:
                    byte_msg = byte_msg.replace(b'\x00', b' ')
                    bflb_utils.printf(byte_msg.decode("GB18030", 'replace'))
                    byte_msg = ''



# -*- coding:utf-8 -*-


import binascii

from ecdsa import ECDH, NIST256p

from libs import bflb_utils


class BflbEcdh(object):

    def __init__(self, curve=NIST256p):
        self.ecdh = ECDH(curve)
        self.local_public_key = None
        self.sharedsecret = ""

    def create_public_key(self):
        self.ecdh.generate_private_key()
        self.local_public_key = self.ecdh.get_public_key()
        ret = binascii.hexlify(self.local_public_key.to_string()).decode("utf-8")
        bflb_utils.printf("local public key:")
        bflb_utils.printf(ret)
        return ret

    def create_shared_key(self, peer_pk):
        self.ecdh.load_received_public_key_bytes(binascii.unhexlify(peer_pk))
        self.sharedsecret = self.ecdh.generate_sharedsecret_bytes()
        ret = binascii.hexlify(self.sharedsecret).decode("utf-8")
        bflb_utils.printf("secret key:")
        bflb_utils.printf(ret)
        return ret



# -*- coding:utf-8 -*-
#  Copyright (C) 2021- BOUFFALO LAB (NANJING) CO., LTD.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.


import os
import sys
import time
import binascii
import struct
import subprocess

import pylink

from libs import bflb_utils
from libs.bflb_utils import app_path

python_version = struct.calcsize("P") * 8

if python_version == 64:
    path_dll = os.path.join(app_path, "JLinkARM_64.dll")
else:
    path_dll = os.path.join(app_path, "JLinkARM_32.dll")


class BflbJLinkPort(object):

    def __init__(self):
        self._speed = 5000
        self._rx_timeout = 10000
        self._jlink_shake_hand_addr = "20000000"
        self._jlink_data_addr = "20000004"
        self._inited = False
        self._chiptype = "bl60x"
        self._jlink_run_addr = "22010000"

    def if_init(self, device, rate, chiptype="bl60x"):
        if self._inited is False:
            # jlink_cfg_module = __import__(chiptype + ".jlink_load_cfg")
            # jlink_cfg_module = __import__("libs." + chiptype + ".jlink_load_cfg", fromlist=["jlink_load_cfg"])
            sub_module = __import__("libs." + chiptype, fromlist=[chiptype])
            self._jlink_shake_hand_addr = sub_module.jlink_load_cfg.jlink_shake_hand_addr
            self._jlink_data_addr = sub_module.jlink_load_cfg.jlink_data_addr
            if sys.platform == 'win32':
                obj_dll = pylink.Library(dllpath=path_dll)
                self._jlink = pylink.JLink(lib=obj_dll)
            else:
                self._jlink = pylink.JLink()
            if device is not None and device.lower().find("com") == -1 and device.lower().find(
                    "tty") == -1:
                bflb_utils.printf(device)
                self._jlink.open(serial_no=int(device))
            else:
                self._jlink.open()
            tif_set = sub_module.jlink_load_cfg.jlink_set_tif
            self._jlink.set_tif(tif_set)
            self._speed = rate
            core_type = sub_module.jlink_load_cfg.jlink_core_type
            self._jlink.connect(core_type, rate)
            self._inited = True
            self._chiptype = chiptype
            self._jlink_run_addr = sub_module.jlink_load_cfg.jlink_run_addr

    def if_set_rx_timeout(self, val):
        self._rx_timeout = val * 1000

    def if_get_rate(self):
        return self._speed

    def halt_cpu(self):
        if self._jlink.halted() is False:
            self._jlink.halt()
        if self._jlink.halted():
            return True
        else:
            bflb_utils.printf("couldn't halt cpu")
            return False

    def reset_cpu(self, ms=0, halt=True):
        self._jlink.reset(ms, halt)

    def set_pc_msp(self, pc, msp):
        if self._jlink.halted() is False:
            self._jlink.halt()
        if self._jlink.halted():
            if self._chiptype == "bl602":
                jlink_script = "jlink.cmd"
                fp = open(jlink_script, 'w+')
                cmd = "h\r\nSetPC " + str(self._jlink_run_addr) + "\r\nexit"
                bflb_utils.printf(cmd)
                fp.write(cmd)
                fp.close()
                # jlink_cmd=r'C:/Keil_v5/ARM/Segger/JLink.exe -device Cortex-M4 -Speed 4000 -IF SWD  -JTAGConf -1,-1 -CommanderScript jlink.cmd'
                jlink_cmd = 'JLink.exe -device RISC-V -Speed ' + str(
                    self._speed
                ) + ' -IF JTAG -jtagconf -1,-1 -autoconnect 1 -CommanderScript jlink.cmd'
                p = subprocess.Popen(jlink_cmd,
                                     shell=True,
                                     stdin=subprocess.PIPE,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)
                out, err = p.communicate()
                bflb_utils.printf(out, err)
                os.remove(jlink_script)
            else:
                self._jlink.register_write(15, int(pc, 16))
                self._jlink.register_write(13, int(msp, 16))
                self._jlink.restart()
        else:
            bflb_utils.printf("couldn't halt cpu")

    def if_write(self, data_send):
        self.if_raw_write(self._jlink_data_addr, data_send)

        #write flag
        data_list = []
        data_list.append(int("59445248", 16))
        self._jlink.memory_write(int(self._jlink_shake_hand_addr, 16), data_list, nbits=32)

    def if_raw_write(self, addr, data_send):
        addr_int = int(addr, 16)
        len2 = len(data_send) % 4
        len1 = len(data_send) - len2
        # using 32bits write
        if len1 != 0:
            data_list = []
            for i in range(int(len1 / 4)):
                data_list.append(data_send[4 * i] + (data_send[4 * i + 1] << 8) +
                                 (data_send[4 * i + 2] << 16) + (data_send[4 * i + 3] << 24))
            self._jlink.memory_write(addr_int, data_list, nbits=32)
        # using 8bits write
        if len2 != 0:
            data_list = []
            for i in range(len2):
                data_list.append(data_send[len1 + i])
            self._jlink.memory_write(addr_int + len1, data_list, nbits=8)

    def if_raw_write8(self, addr, data_send):
        data_list = []
        for data in data_send:
            data_list.append(data)
        self._jlink.memory_write(int(addr, 16), data_list, nbits=8)

    def if_raw_write16(self, addr, data_send):
        data_list = []
        for i in range(int(len(data_send) / 2)):
            data_list.append(data_send[2 * i] + (data_send[2 * i + 1] << 8))
        self._jlink.memory_write(int(addr, 16), data_list, nbits=16)

    def if_raw_write32(self, addr, data_send):
        data_list = []
        for i in range(int(len(data_send) / 4)):
            data_list.append(data_send[4 * i] + (data_send[4 * i + 1] << 8) +
                             (data_send[4 * i + 2] << 16) + (data_send[4 * i + 3] << 24))
        self._jlink.memory_write(int(addr, 16), data_list, nbits=32)

    def if_read(self, data_len):
        start_time = (time.time() * 1000)
        while True:
            ready = self._jlink.memory_read(int(self._jlink_shake_hand_addr, 16), 1, nbits=32)
            # bflb_utils.printf("receiving",ready)
            if ready[0] == int("4B434153", 16):
                break
            elapsed = (time.time() * 1000) - start_time
            # time out judgment
            if elapsed >= self._rx_timeout:
                return 0, "waiting response time out".encode("utf-8")
            time.sleep(0.001)

        data = self.if_raw_read(self._jlink_data_addr, data_len)

        if len(data) != data_len:
            return 0, data
        return 1, data

    def if_raw_read(self, addr, data_len):
        addr_int = int(addr, 16)
        if addr_int % 4 == 0:
            # word align, try to read using nbits=32
            len2 = data_len % 4
            len1 = data_len - len2
            data1 = bytearray(0)
            data2 = bytearray(0)
            if len1 != 0:
                data1 = self._jlink.memory_read(addr_int, int(len1 / 4), nbits=32)
            # for left, using nbits=8
            if len2 != 0:
                data2 = self._jlink.memory_read(addr_int + len1, len2, nbits=8)

            data = bytearray(0)
            for tmp in data1:
                data += bflb_utils.int_to_4bytearray_l(tmp)
            data += bytearray(data2)
            return data
        else:
            return self.if_raw_read8(addr, data_len)

    def if_raw_read8(self, addr, data_len):
        data = self._jlink.memory_read(int(addr, 16), data_len, nbits=8)
        return bytearray(data)

    def if_raw_read16(self, addr, data_len):
        raw_data = self._jlink.memory_read(int(addr, 16), data_len / 2, nbits=16)
        data = bytearray(0)
        for tmp in raw_data:
            data += bflb_utils.int_to_2bytearray_l(tmp)
        return bytearray(data)

    def if_raw_read32(self, addr, data_len):
        raw_data = self._jlink.memory_read(int(addr, 16), data_len / 4, nbits=32)
        data = bytearray(0)
        for tmp in raw_data:
            data += bflb_utils.int_to_4bytearray_l(tmp)
        return bytearray(data)

    def if_shakehand(self,
                     do_reset=False,
                     reset_hold_time=100,
                     shake_hand_delay=100,
                     reset_revert=True):
        self.if_write(bytearray(1))
        success, ack = self.if_read(2)
        bflb_utils.printf(binascii.hexlify(ack))
        if ack.find(b'\x4F') != -1 or ack.find(b'\x4B') != -1:
            time.sleep(0.03)
            return "OK"
        return "FL"

    def if_close(self):
        self._jlink.close()
        self._inited = False

    def if_deal_ack(self):
        success, ack = self.if_read(2)
        if success == 0:
            bflb_utils.printf("ack:" + str(binascii.hexlify(ack)))
            return ack.decode("utf-8")
        if ack.find(b'\x4F') != -1 or ack.find(b'\x4B') != -1:
            return "OK"
        success, err_code = self.if_read(4)
        if success == 0:
            bflb_utils.printf("err_code:" + str(binascii.hexlify(err_code)))
            return "FL"
        err_code_str = str(binascii.hexlify(err_code[3:4] + err_code[2:3]).decode('utf-8'))
        ack = "FL"
        try:
            ret = ack + err_code_str + "(" + bflb_utils.get_bflb_error_code[err_code_str] + ")"
        except Exception:
            ret = ack + err_code_str + " unknown"
        bflb_utils.printf(ret)
        return ret

    def if_deal_response(self):
        ack = self.if_deal_ack()
        if ack == "OK":
            success, len_bytes = self.if_read(4)
            if success == 0:
                bflb_utils.printf("Get length error")
                bflb_utils.printf(binascii.hexlify(len_bytes))
                return "Get length error", len_bytes
            # byte 2 and 3 is length
            tmp = bflb_utils.bytearray_reverse(len_bytes[2:4])
            data_len = bflb_utils.bytearray_to_int(tmp)
            success, data_bytes = self.if_read(data_len + 4)
            if success == 0:
                bflb_utils.printf("Read data error")
                return "Read data error", data_bytes
            # remove thr first 4 bytes
            data_bytes = data_bytes[4:]
            if len(data_bytes) != data_len:
                bflb_utils.printf("Not get excepted length")
                return "Not get excepted length", data_bytes
            return ack, data_bytes
        bflb_utils.printf("Not ack OK")
        bflb_utils.printf(ack)
        return ack, None



# -*- coding: utf-8 -*-
#  Copyright (C) 2021- BOUFFALO LAB (NANJING) CO., LTD.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.


import re
import os
import sys
import time
import shutil
import zipfile

try:
    import bflb_path
except ImportError:
    from libs import bflb_path
from libs import bflb_utils
from libs import bflb_efuse_boothd_create
from libs.bflb_utils import app_path, set_error_code
from libs.bflb_configobj import BFConfigParser


def take_second(elem):
    return elem[1]


def factory_mode_set(file, value):
    cfg = BFConfigParser()
    cfg.read(file)
    if cfg.has_option("EFUSE_CFG", "factory_mode"):
        cfg.set("EFUSE_CFG", "factory_mode", value)
        cfg.write(file, 'w')


def check_pt_file(file, addr):
    if len(file) > 0:
        i = 0
        L = []
        while i < len(file):
            L.append([file[i].replace("\\", "/"), int(addr[i], 16)])
            i += 1
        L.sort(key=take_second)
        i = 0
        try:
            while i < len(L) - 1:
                address = L[i][1]
                address_next = L[i + 1][1]
                file_size = os.path.getsize(os.path.join(app_path, L[i][0]))
                if address_next < address + file_size:
                    bflb_utils.printf(
                        "pt check fail, %s is overlayed with %s in flash layout, please check your partition table to fix this issue"
                        % (L[i][0], L[i + 1][0]))
                    return False
                i += 1
        except Exception as e:
            bflb_utils.printf(e)
            return False
    return True


def compress_dir(chipname, zippath, efuse_load=False):
    zip_file = os.path.join(app_path, chipname, zippath, "whole_img.pack")
    dir_path = os.path.join(app_path, chipname, chipname)
    cfg_file = os.path.join(app_path, chipname, "eflash_loader/eflash_loader_cfg.ini")
    cfg = BFConfigParser()
    cfg.read(cfg_file)
    flash_file = re.compile('\s+').split(cfg.get("FLASH_CFG", "file"))
    address = re.compile('\s+').split(cfg.get("FLASH_CFG", "address"))
    if check_pt_file(flash_file, address) is not True:
        bflb_utils.printf("PT Check Fail")
        set_error_code("0082")
        return False
    factory_mode_set(os.path.join(app_path, chipname, "eflash_loader/eflash_loader_cfg.ini"), "true")
    flash_file.append(os.path.join(app_path, chipname, "eflash_loader/eflash_loader_cfg.ini"))
    if efuse_load:
        flash_file.append(cfg.get("EFUSE_CFG", "file"))
        flash_file.append(cfg.get("EFUSE_CFG", "maskfile"))
    if len(flash_file) > 0:
        i = 0
        try:
            while i < len(flash_file):
                relpath = os.path.relpath(flash_file[i].replace("\\", "/"), app_path)
                dir = os.path.join(app_path, chipname, relpath)
                if os.path.isdir(os.path.dirname(dir)) is False:
                    os.makedirs(os.path.dirname(dir))
                shutil.copy(os.path.join(app_path, flash_file[i].replace("\\", "/")), dir)
                i += 1
            verfile = os.path.join(app_path, chipname, chipname, "version.txt")
            with open(verfile, mode="w") as f:
                f.write(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
        except Exception as e:
            bflb_utils.printf(e)
            factory_mode_set(os.path.join(chipname, "eflash_loader/eflash_loader_cfg.ini"),
                             "false")
            return False

    try:
        z = zipfile.ZipFile(zip_file, 'w')
        for dirpath, dirnames, filenames in os.walk(dir_path):
            for file in filenames:
                # z.write(os.path.relpath(os.path.join(dirpath, file), os.path.join(app_path, chipname)))
                z.write(
                    os.path.join(dirpath, file),
                    os.path.relpath(os.path.join(dirpath, file), os.path.join(app_path, chipname)))
        z.close()
        shutil.rmtree(dir_path)
    except Exception as e:
        bflb_utils.printf(e)
        factory_mode_set(os.path.join(chipname, "eflash_loader/eflash_loader_cfg.ini"), "false")
        return False
    factory_mode_set(os.path.join(chipname, "eflash_loader/eflash_loader_cfg.ini"), "false")
    return True


def img_create(args, chipname="bl60x", chiptype="bl60x", img_dir=None, config_file=None):
    sub_module = __import__("libs." + chiptype, fromlist=[chiptype])
    img_dir_path = os.path.join(app_path, chipname, "img_create_iot")
    if img_dir is None:
        sub_module.img_create_do.img_create_do(args, img_dir_path, config_file)
    else:
        sub_module.img_create_do.img_create_do(args, img_dir, config_file)


def create_sp_media_image_file(config, chiptype="bl60x", cpu_type=None):
    sub_module = __import__("libs." + chiptype, fromlist=[chiptype])
    sub_module.img_create_do.create_sp_media_image(config, cpu_type)


def encrypt_loader_bin(chiptype, file, sign, encrypt, createcfg):
    sub_module = __import__("libs." + chiptype, fromlist=[chiptype])
    return sub_module.img_create_do.encrypt_loader_bin_do(file, sign, encrypt, createcfg)


def run():
    parser_image = bflb_utils.image_create_parser_init()
    args = parser_image.parse_args()
    # args = parser_image.parse_args("--image=media", "--signer=none")
    bflb_utils.printf("Chipname: %s" % args.chipname)
    if args.chipname:
        chip_dict = {
            "bl56x": "bl60x",
            "bl60x": "bl60x",
            "bl562": "bl602",
            "bl602": "bl602",
            "bl702": "bl702",
            "bl606p": "bl606p",
        }
        chipname = args.chipname
        chiptype = chip_dict[chipname]
        img_create_path = os.path.join(app_path, chipname, "img_create_mcu")
        img_create_cfg = os.path.join(app_path, chipname, "img_create_mcu") + "/img_create_cfg.ini"
        bh_cfg_file = img_create_path + "/bootheader_cfg.ini"
        bh_file = img_create_path + "/bootheader.bin"
        if args.imgfile:
            imgbin = args.imgfile
            cfg = BFConfigParser()
            cfg.read(img_create_cfg)
            cfg.set('Img_Cfg', 'segdata_file', imgbin)
            cfg.write(img_create_cfg, 'w')
        bflb_efuse_boothd_create.bootheader_create_process(chipname, chiptype, bh_cfg_file, bh_file,
            img_create_path + "/bootheader_dummy.bin")
        img_create(args, chipname, chiptype, img_create_path, img_create_cfg)
    else:
        bflb_utils.printf("Please set chipname config, exit")


if __name__ == '__main__':
    run()




# -*- coding:utf-8 -*-


import binascii
from libs import bflb_utils

partition1_addr = "E000"
partition2_addr = "F000"
fireware_name = "FW_CPU0"
partition_magic_code = 0x42465054


def check_pt_data(data):
    """
    partition data 0~15 is partition table config
    parrition data 12~15 is partition table config crc32
    partition data 16~16+36 is partition entry 1, 16+36~16+72 is partition entry 2 ...
    partition data last 4 byte is partition entry data crc32
    """
    # bflb_utils.printf(binascii.hexlify(data))
    if partition_magic_code != bflb_utils.bytearray_to_int(data[0:4]):
        bflb_utils.printf("partition bin magic check fail ", binascii.hexlify(data[0:4]))
        return False, 0, 0
    table_count = bflb_utils.bytearray_to_int(
        data[6:7]) + (bflb_utils.bytearray_to_int(data[7:8]) << 8)
    # bflb_utils.printf("table count: ", table_count)
    if table_count > 16:
        bflb_utils.printf("error, pt enter size > 16")
        return False, 0, 0
    crcarray = bflb_utils.get_crc32_bytearray(data[:12])
    if data[12:16] != crcarray:
        bflb_utils.printf("pt table crc fail ", binascii.hexlify(crcarray))
        return False, 0, 0
    crcarray = bflb_utils.get_crc32_bytearray(data[16:16 + (36 * table_count)])
    if data[16 + (36 * table_count):16 + (36 * table_count) + 4] != crcarray:
        bflb_utils.printf("pt entries crc fail ", binascii.hexlify(crcarray))
        return False, 0, 0
    age = bflb_utils.bytearray_to_int(data[8:9]) + (bflb_utils.bytearray_to_int(data[9:10])<<8) +\
         (bflb_utils.bytearray_to_int(data[10:11])<<16) + (bflb_utils.bytearray_to_int(data[11:12])<<24)
    return True, table_count, age



# -*- coding:utf-8 -*-

from . import flash_select_do
from . import img_create_do
from . import bootheader_cfg_keys
from . import efuse_cfg_keys
from . import efuse_create_do
from . import efuse_data_create
from . import jlink_load_cfg
from . import openocd_load_cfg
from . import partition_cfg_do



# -*- coding: utf-8 -*-


import os

from libs import bflb_utils
from libs.bflb_utils import app_path
from libs.bflb_configobj import BFConfigParser


def get_suitable_file_name(cfg_dir, flash_id):
    conf_files = []
    for home, dirs, files in os.walk(cfg_dir):
        for filename in files:
            if filename.split('_')[-1] == flash_id + '.conf':
                conf_files.append(filename)

    if len(conf_files) > 1:
        bflb_utils.printf("Flash id duplicate and alternative is:")
        for i in range(len(conf_files)):
            tmp = conf_files[i].split('.')[0]
            bflb_utils.printf("%d:%s" % (i + 1, tmp))
        return conf_files[i]
    elif len(conf_files) == 1:
        return conf_files[0]
    else:
        return ""


def update_flash_cfg_do(chipname, chiptype, flash_id, file=None, create=False, section=None):
    cfg_dir = app_path + "/common/flash_config/" + chiptype + '/'
    conf_name = get_suitable_file_name(cfg_dir, flash_id)
    value_key = []
    if os.path.isfile(cfg_dir + conf_name) is False:
        return False
    fp = open(cfg_dir + conf_name, 'r')
    for line in fp.readlines():
        value = line.split("=")[0].strip()
        if value == "[FLASH_CFG]":
            continue
        value_key.append(value)

    cfg1 = BFConfigParser()
    cfg1.read(cfg_dir + conf_name)
    cfg2 = BFConfigParser()
    cfg2.read(file)
    for i in range(len(value_key)):
        if cfg1.has_option("FLASH_CFG", value_key[i]) and cfg2.has_option(section, value_key[i]):
            tmp_value = cfg1.get("FLASH_CFG", value_key[i])
            bflb_utils.update_cfg(cfg2, section, value_key[i], tmp_value)

    cfg2.write(file, "w+")

    bflb_utils.printf("Update flash cfg finished")


def get_supported_flash_do():
    flash_type = []
    #for itr in flashList:
    #    flash_type.append(itr["name"])
    return flash_type



# -*- coding: utf-8 -*-
#  Copyright (C) 2021- BOUFFALO LAB (NANJING) CO., LTD.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.


from libs import bflb_utils
from libs import bflb_toml as toml


class PtCreater(object):

    def __init__(self, config_file):
        self.parsed_toml = toml.load(config_file)
        self.entry_max = 16
        self.pt_new = False

    def __create_pt_table_do(self, lists, file):
        entry_table = bytearray(36 * self.entry_max)
        entry_cnt = 0
        for item in lists:
            entry_type = item["type"]
            entry_name = item["name"]
            entry_device = item["device"]
            entry_addr0 = item["address0"]
            entry_addr1 = item["address1"]
            entry_maxlen0 = item["size0"]
            entry_maxlen1 = item["size1"]
            entry_len = item["len"]
            entry_table[36 * entry_cnt + 0] = bflb_utils.int_to_2bytearray_l(entry_type)[0]
            if "activeindex" in item:
                entry_activeindex = item["activeindex"]
                entry_table[36 * entry_cnt +
                            2] = bflb_utils.int_to_2bytearray_l(entry_activeindex)[0]
            if len(entry_name) >= 8:
                bflb_utils.printf("Entry name is too long!")
                return False
            entry_table[36 * entry_cnt + 3:36 * entry_cnt + 3 +
                        len(entry_name)] = bytearray(entry_name, "utf-8") + bytearray(0)
            entry_table[36 * entry_cnt + 12:36 * entry_cnt +
                        16] = bflb_utils.int_to_4bytearray_l(entry_addr0)
            entry_table[36 * entry_cnt + 16:36 * entry_cnt +
                        20] = bflb_utils.int_to_4bytearray_l(entry_addr1)
            entry_table[36 * entry_cnt + 20:36 * entry_cnt +
                        24] = bflb_utils.int_to_4bytearray_l(entry_maxlen0)
            entry_table[36 * entry_cnt + 24:36 * entry_cnt +
                        28] = bflb_utils.int_to_4bytearray_l(entry_maxlen1)
            if "age" in item:
                entry_age = item["age"]
                entry_table[36 * entry_cnt + 32:36 * entry_cnt +
                            36] = bflb_utils.int_to_4bytearray_l(entry_age)
            entry_cnt += 1
        # partition table header
        # 0x54504642
        pt_table = bytearray(16)
        pt_table[0] = 0x42
        pt_table[1] = 0x46
        pt_table[2] = 0x50
        pt_table[3] = 0x54
        pt_table[6:8] = bflb_utils.int_to_2bytearray_l(int(entry_cnt))
        pt_table[12:16] = bflb_utils.get_crc32_bytearray(pt_table[0:12])
        entry_table[36 * entry_cnt:36 * entry_cnt + 4] = bflb_utils.get_crc32_bytearray(
            entry_table[0:36 * entry_cnt])
        data = pt_table + entry_table[0:36 * entry_cnt + 4]
        fp = open(file, 'wb+')
        fp.write(data)
        fp.close()
        return True

    def create_pt_table(self, file):
        self.pt_new = True
        return self.__create_pt_table_do(self.parsed_toml["pt_entry"], file)

    def get_pt_table_addr(self):
        addr0 = self.parsed_toml["pt_table"]["address0"]
        addr1 = self.parsed_toml["pt_table"]["address1"]
        return addr0, addr1

    def construct_table(self):
        parcel = {}
        if self.pt_new is True:
            parcel['pt_new'] = True
        else:
            parcel['pt_new'] = False
        parcel['pt_addr0'] = self.parsed_toml["pt_table"]["address0"]
        parcel['pt_addr1'] = self.parsed_toml["pt_table"]["address1"]
        for tbl_item in self.parsed_toml["pt_entry"]:
            if tbl_item['name'] == 'factory':
                parcel['conf_addr'] = tbl_item['address0']
            if tbl_item['name'] == 'FW_CPU0':
                parcel['fw_cpu0_addr'] = tbl_item['address0']
            if tbl_item['name'] == 'FW':
                parcel['fw_addr'] = tbl_item['address0']
            if tbl_item['name'] == 'media':
                parcel['media_addr'] = tbl_item['address0']
            if tbl_item['name'] == 'mfg':
                parcel['mfg_addr'] = tbl_item['address0']
        return parcel


if __name__ == '__main__':
    pt_helper = PtCreater("partition_cfg.toml")
    pt_helper.create_pt_table("partition_test.bin")
    pt_helper.get_pt_table_addr()



# -*- coding: utf-8 -*-
#  Copyright (C) 2021- BOUFFALO LAB (NANJING) CO., LTD.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

import os
import re

from libs import bflb_utils
from libs.bflb_utils import app_path
from libs.bflb_configobj import BFConfigParser


def get_int_mask(pos, length):
    ones = "1" * 32
    zeros = "0" * 32
    mask = ones[0:32 - pos - length] + zeros[0:length] + ones[0:pos]
    return int(mask, 2)


def update_data_from_cfg(config_keys, config_file):
    section = "FLASH_CFG"
    # bflb_utils.printf("Updating data according to <" + config_file + "[" + section + "]>")
    cfg = BFConfigParser()
    cfg.read(config_file)
    # get finally data len
    filelen = 0
    offset = 0
    minOffset = 0xFFFFFFFF
    maxOffset = 0
    flashCrcOffset = 0
    crcOffset = 0
    if config_keys.get("crc32") != None:
        crcOffset = int(config_keys.get("crc32")["offset"], 10)
    if config_keys.get("flashcfg_crc32") != None:
        flashCrcOffset = int(config_keys.get("flashcfg_crc32")["offset"], 10)
    for key in cfg.options(section):
        if config_keys.get(key) == None:
            continue
        offset = int(config_keys.get(key)["offset"], 10)
        if offset < minOffset:
            minOffset = offset
        if offset > maxOffset:
            maxOffset = offset
    filelen = maxOffset - minOffset + 4
    data = bytearray(filelen)
    # bflb_utils.printf(binascii.hexlify(data))
    for key in cfg.options(section):
        if config_keys.get(key) == None:
            bflb_utils.printf(key + " not exist")
            continue
        # bflb_utils.printf(key)
        val = cfg.get(section, key)
        if val.startswith("0x"):
            val = int(val, 16)
        else:
            val = int(val, 10)
        # bflb_utils.printf(val)
        offset = int(config_keys.get(key)["offset"], 10) - minOffset
        pos = int(config_keys.get(key)["pos"], 10)
        bitlen = int(config_keys.get(key)["bitlen"], 10)

        oldval = bflb_utils.bytearray_to_int(bflb_utils.bytearray_reverse(data[offset:offset + 4]))
        newval = (oldval & get_int_mask(pos, bitlen)) + (val << pos)
        # bflb_utils.printf(newval,binascii.hexlify(bflb_utils.int_to_4bytearray_l(newval)))
        data[offset:offset + 4] = bflb_utils.int_to_4bytearray_l(newval)
    # bflb_utils.printf(binascii.hexlify(data))
    return minOffset, filelen, data, flashCrcOffset, crcOffset


def update_flash_cfg_data_do(chipname, chiptype, flash_id):
    cfg_dir = app_path + "/common/flash_config/" + chiptype + '/'
    sub_module = __import__("libs." + chiptype, fromlist=[chiptype])
    conf_name = sub_module.flash_select_do.get_suitable_file_name(cfg_dir, flash_id)
    if os.path.isfile(cfg_dir + conf_name) == False:
        return False
    return update_data_from_cfg(sub_module.bootheader_cfg_keys.bootheader_cfg_keys, cfg_dir + conf_name)


def flash_bootheader_config_check(chipname, chiptype, flashid, file):
    magic_code = 0x504e4642
    flash_magic_code = 0x47464346
    offset, flashCfgLen, data, flashCrcOffset, crcOffset =\
        update_flash_cfg_data_do(chipname, chiptype, flashid)

    fp = open(file, 'rb')
    rdata = bytearray(fp.read())
    fp.close()
    i = 0
    length = 128
    while i < length:
        if rdata[i:i + 4] == bflb_utils.int_to_4bytearray_l(magic_code):
            if rdata[i + 8:i + 12] == bflb_utils.int_to_4bytearray_l(flash_magic_code):
                data[2:4] = rdata[i + 14:i + 16]
                flashCfg = rdata[i + offset:i + offset + flashCfgLen]
                if data != flashCfg:
                    return False
        i += 4
    return True


def update_flash_cfg_data(chipname, chiptype, flash_id, cfg, bh_cfg_file, cfg_key):
    cfg2 = BFConfigParser()
    cfg2.read(bh_cfg_file)
    magic_code = cfg2.get(cfg_key, "magic_code")
    magic_code = int(magic_code, 16)
    flash_magic_code = cfg2.get(cfg_key, "flashcfg_magic_code")
    flash_magic_code = int(flash_magic_code, 16)
    sub_module = __import__("libs." + chiptype, fromlist=[chiptype])
    offset, flashCfgLen, data, flashCrcOffset, crcOffset = update_flash_cfg_data_do(chipname, chiptype, flash_id)

    para_file = cfg.get("FLASH_CFG", "flash_para")
    fp = open(para_file, 'wb')
    fp.write(data)
    fp.close()

    flash_file = re.compile('\s+').split(cfg.get("FLASH_CFG", "file"))
    for f in flash_file:
        fp = open(f, 'rb')
        rdata = bytearray(fp.read())
        fp.close()
        i = 0
        length = len(rdata)
        while i < length:
            if rdata[i:i + 4] == bflb_utils.int_to_4bytearray_l(magic_code):
                if rdata[i + 8:i + 12] == bflb_utils.int_to_4bytearray_l(flash_magic_code):
                    data[2:4] = rdata[i + 14:i + 16]
                    flashCfg = rdata[i + offset:i + offset + flashCfgLen]
                    if data != flashCfg:
                        return False
            i += 4
    return True


def update_flash_cfg(chipname, chiptype, flash_id, file=None, create=False, section=None):
    sub_module = __import__("libs." + chiptype, fromlist=[chiptype])
    if sub_module.flash_select_do.update_flash_cfg_do(chipname, chiptype, flash_id, file, create, section) == False:
        return False
    return True


def get_supported_flash(chiptype):
    sub_module = __import__("libs." + chiptype, fromlist=[chiptype])
    return sub_module.flash_select_do.get_supported_flash_do()





# -*- coding:utf-8 -*-


import hashlib
import binascii

import ecdsa

from libs import bflb_utils
from libs.bflb_utils import open_file, img_create_sha256_data, img_create_encrypt_data
from libs.bflb_configobj import BFConfigParser


keyslot0 = 28
keyslot1 = keyslot0 + 16
keyslot2 = keyslot1 + 16
keyslot3 = keyslot2 + 16
keyslot4 = keyslot3 + 16
keyslot5 = keyslot4 + 16
keyslot6 = 156
keyslot7 = keyslot6 + 16
keyslot8 = keyslot7 + 16
keyslot9 = keyslot8 + 16
keyslot10 = keyslot9 + 16
keyslot11 = keyslot10 + 16
keyslot12 = keyslot11 + 16

rd_lock_key_slot_5 = 31
rd_lock_key_slot_4 = 30
rd_lock_key_slot_3 = 28
rd_lock_key_slot_2 = 28
rd_lock_key_slot_1 = 27
rd_lock_key_slot_0 = 26
rd_lock_dbg_pwd = 23

wr_lock_key_slot_5 = 15
wr_lock_key_slot_4 = 14
wr_lock_key_slot_3 = 13
wr_lock_key_slot_2 = 12
wr_lock_key_slot_1 = 11
wr_lock_key_slot_0 = 10
wr_lock_wifi_mac = 9
wr_lock_chip_id = 8
wr_lock_dbg_pwd = 7
wr_lock_sf_aes_mode = 6
wr_lock_sboot_sign_mode = 5

rd_lock_key_slot_11 = 31
rd_lock_key_slot_10 = 30
rd_lock_key_slot_9 = 29
rd_lock_key_slot_8 = 28
rd_lock_key_slot_7 = 27
rd_lock_key_slot_6 = 26

wr_lock_key_slot_11 = 15
wr_lock_key_slot_10 = 14
wr_lock_key_slot_9 = 13
wr_lock_key_slot_8 = 12
wr_lock_key_slot_7 = 11
wr_lock_key_slot_6 = 10
wr_lock_sw_usage_3 = 9
wr_lock_sw_usage_2 = 8
wr_lock_sw_usage_1 = 7
wr_lock_sw_usage_0 = 6
wr_lock_ana_trim_2 = 5
wr_lock_ana_trim_1 = 4
wr_lock_ana_trim_0 = 3


# update efuse info
def img_update_efuse_sp(cfg, sign, pk_hash, flash_encryp_type, flash_key, sec_eng_key_sel,
                        sec_eng_key):
    fp = open_file(cfg.get("Img_Cfg", "efuse_file"), 'rb')
    efuse_data = bytearray(fp.read()) + bytearray(0)
    fp.close()
    fp = open_file(cfg.get("Img_Cfg", "efuse_mask_file"), 'rb')
    efuse_mask_data = bytearray(fp.read()) + bytearray(0)
    fp.close()
    mask_4bytes = bytearray.fromhex("FFFFFFFF")
    efuse_data[0] |= flash_encryp_type
    efuse_data[0] |= (sign << 2)
    efuse_mask_data[0] |= 0xf
    rw_lock0 = 0
    rw_lock1 = 0
    if pk_hash is not None:
        efuse_data[keyslot0:keyslot2] = pk_hash
        efuse_mask_data[keyslot0:keyslot2] = mask_4bytes * 8
        rw_lock0 |= (1 << wr_lock_key_slot_0)
        rw_lock0 |= (1 << wr_lock_key_slot_1)
    if flash_key is not None:
        efuse_data[keyslot2:keyslot4] = flash_key
        efuse_mask_data[keyslot2:keyslot4] = mask_4bytes * 8
        rw_lock0 |= (1 << wr_lock_key_slot_2)
        rw_lock0 |= (1 << wr_lock_key_slot_3)
        rw_lock0 |= (1 << rd_lock_key_slot_2)
        rw_lock0 |= (1 << rd_lock_key_slot_3)
    if sec_eng_key is not None:
        if flash_encryp_type == 0:
            if sec_eng_key_sel == 0:
                efuse_data[keyslot2:keyslot3] = sec_eng_key[16:32]
                efuse_data[keyslot3:keyslot4] = sec_eng_key[0:16]
                efuse_mask_data[keyslot2:keyslot4] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_2)
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_2)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
            if sec_eng_key_sel == 1:
                efuse_data[keyslot3:keyslot4] = sec_eng_key[16:32]
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot3:keyslot5] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 2:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[16:32]
                efuse_data[keyslot7:keyslot8] = sec_eng_key[0:16]
                efuse_mask_data[keyslot4:keyslot5] = mask_4bytes * 4
                efuse_mask_data[keyslot7:keyslot8] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock1 |= (1 << wr_lock_key_slot_7)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
                rw_lock1 |= (1 << rd_lock_key_slot_7)
            if sec_eng_key_sel == 3:
                efuse_data[keyslot7:keyslot8] = sec_eng_key[16:32]
                efuse_data[keyslot2:keyslot3] = sec_eng_key[0:16]
                efuse_mask_data[keyslot7:keyslot8] = mask_4bytes * 4
                efuse_mask_data[keyslot2:keyslot3] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_2)
                rw_lock1 |= (1 << wr_lock_key_slot_7)
                rw_lock0 |= (1 << rd_lock_key_slot_2)
                rw_lock1 |= (1 << rd_lock_key_slot_7)
        if flash_encryp_type == 1:
            if sec_eng_key_sel == 0:
                efuse_data[keyslot3:keyslot4] = sec_eng_key[16:32]
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot3:keyslot5] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 1:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[16:32]
                efuse_data[keyslot3:keyslot4] = sec_eng_key[0:16]
                efuse_mask_data[keyslot3:keyslot5] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 2:
                efuse_data[keyslot3:keyslot4] = sec_eng_key[16:32]
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot3:keyslot5] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 3:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[16:32]
                efuse_data[keyslot3:keyslot4] = sec_eng_key[0:16]
                efuse_mask_data[keyslot3:keyslot5] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
        if flash_encryp_type == 2 or flash_encryp_type == 3:
            if sec_eng_key_sel == 0:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot4:keyslot5] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 1:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot4:keyslot5] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 2:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot4:keyslot5] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 3:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot4:keyslot5] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
    efuse_data[124:128] = bflb_utils.int_to_4bytearray_l(rw_lock0)
    efuse_mask_data[124:128] = bflb_utils.int_to_4bytearray_l(rw_lock0)
    efuse_data[252:256] = bflb_utils.int_to_4bytearray_l(rw_lock1)
    efuse_mask_data[252:256] = bflb_utils.int_to_4bytearray_l(rw_lock1)
    fp = open_file(cfg.get("Img_Cfg", "efuse_file"), 'wb+')
    fp.write(efuse_data)
    fp.close()
    fp = open_file(cfg.get("Img_Cfg", "efuse_mask_file"), 'wb+')
    fp.write(efuse_mask_data)
    fp.close()


def img_update_efuse_cpu0(cfg, sign, pk_hash, flash_encryp_type, flash_key, sec_eng_key_sel,
                          sec_eng_key):
    fp = open_file(cfg.get("Img_CPU0_Cfg", "efuse_file"), 'rb')
    efuse_data = bytearray(fp.read()) + bytearray(0)
    fp.close()
    fp = open_file(cfg.get("Img_CPU0_Cfg", "efuse_mask_file"), 'rb')
    efuse_mask_data = bytearray(fp.read()) + bytearray(0)
    fp.close()
    mask_4bytes = bytearray.fromhex("FFFFFFFF")
    efuse_data[0] |= flash_encryp_type
    efuse_data[0] |= (sign << 2)
    efuse_mask_data[0] |= 0xf
    rw_lock0 = 0
    rw_lock1 = 0
    if pk_hash is not None:
        efuse_data[keyslot0:keyslot2] = pk_hash
        efuse_mask_data[keyslot0:keyslot2] = mask_4bytes * 8
        rw_lock0 |= (1 << wr_lock_key_slot_0)
        rw_lock0 |= (1 << wr_lock_key_slot_1)
    if flash_key is not None:
        efuse_data[keyslot2:keyslot4] = flash_key
        efuse_mask_data[keyslot2:keyslot4] = mask_4bytes * 8
        rw_lock0 |= (1 << wr_lock_key_slot_2)
        rw_lock0 |= (1 << wr_lock_key_slot_3)
        rw_lock0 |= (1 << rd_lock_key_slot_2)
        rw_lock0 |= (1 << rd_lock_key_slot_3)
    if sec_eng_key is not None:
        if flash_encryp_type == 0:
            if sec_eng_key_sel == 0:
                efuse_data[keyslot2:keyslot3] = sec_eng_key[16:32]
                efuse_data[keyslot3:keyslot4] = sec_eng_key[0:16]
                efuse_mask_data[keyslot2:keyslot4] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_2)
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_2)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
            if sec_eng_key_sel == 1:
                efuse_data[keyslot3:keyslot4] = sec_eng_key[16:32]
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot3:keyslot5] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 2:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[16:32]
                efuse_data[keyslot7:keyslot8] = sec_eng_key[0:16]
                efuse_mask_data[keyslot4:keyslot5] = mask_4bytes * 4
                efuse_mask_data[keyslot7:keyslot8] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock1 |= (1 << wr_lock_key_slot_7)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
                rw_lock1 |= (1 << rd_lock_key_slot_7)
            if sec_eng_key_sel == 3:
                efuse_data[keyslot7:keyslot8] = sec_eng_key[16:32]
                efuse_data[keyslot2:keyslot3] = sec_eng_key[0:16]
                efuse_mask_data[keyslot7:keyslot8] = mask_4bytes * 4
                efuse_mask_data[keyslot2:keyslot3] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_2)
                rw_lock1 |= (1 << wr_lock_key_slot_7)
                rw_lock0 |= (1 << rd_lock_key_slot_2)
                rw_lock1 |= (1 << rd_lock_key_slot_7)
        if flash_encryp_type == 1:
            if sec_eng_key_sel == 0:
                efuse_data[keyslot3:keyslot4] = sec_eng_key[16:32]
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot3:keyslot5] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 1:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[16:32]
                efuse_data[keyslot3:keyslot4] = sec_eng_key[0:16]
                efuse_mask_data[keyslot3:keyslot5] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 2:
                efuse_data[keyslot3:keyslot4] = sec_eng_key[16:32]
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot3:keyslot5] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 3:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[16:32]
                efuse_data[keyslot3:keyslot4] = sec_eng_key[0:16]
                efuse_mask_data[keyslot3:keyslot5] = mask_4bytes * 8
                rw_lock0 |= (1 << wr_lock_key_slot_3)
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_3)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
        if flash_encryp_type == 2 or flash_encryp_type == 3:
            if sec_eng_key_sel == 0:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot4:keyslot5] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 1:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot4:keyslot5] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 2:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot4:keyslot5] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
            if sec_eng_key_sel == 3:
                efuse_data[keyslot4:keyslot5] = sec_eng_key[0:16]
                efuse_mask_data[keyslot4:keyslot5] = mask_4bytes * 4
                rw_lock0 |= (1 << wr_lock_key_slot_4)
                rw_lock0 |= (1 << rd_lock_key_slot_4)
    efuse_data[124:128] = bflb_utils.int_to_4bytearray_l(rw_lock0)
    efuse_mask_data[124:128] = bflb_utils.int_to_4bytearray_l(rw_lock0)
    efuse_data[252:256] = bflb_utils.int_to_4bytearray_l(rw_lock1)
    efuse_mask_data[252:256] = bflb_utils.int_to_4bytearray_l(rw_lock1)
    fp = open_file(cfg.get("Img_CPU0_Cfg", "efuse_file"), 'wb+')
    fp.write(efuse_data)
    fp.close()
    fp = open_file(cfg.get("Img_CPU0_Cfg", "efuse_mask_file"), 'wb+')
    fp.write(efuse_mask_data)
    fp.close()


def img_update_efuse_cpu1(cfg, sign, pk_hash, flash_encryp_type, flash_key, sec_eng_key_sel,
                          sec_eng_key):
    fp = open_file(cfg.get("Img_CPU1_Cfg", "efuse_file"), 'rb')
    efuse_data = bytearray(fp.read()) + bytearray(0)
    fp.close()
    fp = open_file(cfg.get("Img_CPU1_Cfg", "efuse_mask_file"), 'rb')
    efuse_mask_data = bytearray(fp.read()) + bytearray(0)
    fp.close()
    mask_4bytes = bytearray.fromhex("FFFFFFFF")
    efuse_data[0] |= flash_encryp_type
    efuse_data[0] |= (sign << 2)
    efuse_mask_data[0] |= 0xf
    rw_lock0 = 0
    rw_lock1 = 0
    if pk_hash is not None:
        efuse_data[keyslot5:keyslot5 + 16] = pk_hash[0:16]
        efuse_mask_data[keyslot5:keyslot5 + 16] = mask_4bytes * 4
        efuse_data[keyslot6:keyslot7] = pk_hash[16:32]
        efuse_mask_data[keyslot6:keyslot7] = mask_4bytes * 4
        rw_lock0 |= (1 << wr_lock_key_slot_5)
        rw_lock1 |= (1 << wr_lock_key_slot_6)
    if flash_key is not None:
        efuse_data[keyslot7:keyslot9] = flash_key
        efuse_mask_data[keyslot7:keyslot9] = mask_4bytes * 8
        rw_lock1 |= (1 << wr_lock_key_slot_7)
        rw_lock1 |= (1 << wr_lock_key_slot_8)
        rw_lock1 |= (1 << rd_lock_key_slot_7)
        rw_lock1 |= (1 << rd_lock_key_slot_8)
    if sec_eng_key is not None:
        if flash_encryp_type == 0:
            if sec_eng_key_sel == 0:
                efuse_data[keyslot8:keyslot9] = sec_eng_key[16:32]
                efuse_data[keyslot9:keyslot10] = sec_eng_key[0:16]
                efuse_mask_data[keyslot8:keyslot10] = mask_4bytes * 8
                rw_lock1 |= (1 << wr_lock_key_slot_8)
                rw_lock1 |= (1 << wr_lock_key_slot_9)
                rw_lock1 |= (1 << rd_lock_key_slot_8)
                rw_lock1 |= (1 << rd_lock_key_slot_9)
            if sec_eng_key_sel == 1:
                efuse_data[keyslot9:keyslot10] = sec_eng_key[16:32]
                efuse_data[keyslot10:keyslot11] = sec_eng_key[0:16]
                efuse_mask_data[keyslot9:keyslot11] = mask_4bytes * 8
                rw_lock1 |= (1 << wr_lock_key_slot_9)
                rw_lock1 |= (1 << wr_lock_key_slot_10)
                rw_lock1 |= (1 << rd_lock_key_slot_9)
                rw_lock1 |= (1 << rd_lock_key_slot_10)
            if sec_eng_key_sel == 2:
                efuse_data[keyslot10:keyslot11] = sec_eng_key[16:32]
                efuse_data[keyslot11:keyslot12] = sec_eng_key[0:16]
                efuse_mask_data[keyslot10:keyslot12] = mask_4bytes * 8
                rw_lock1 |= (1 << wr_lock_key_slot_10)
                rw_lock1 |= (1 << wr_lock_key_slot_11)
                rw_lock1 |= (1 << rd_lock_key_slot_10)
                rw_lock1 |= (1 << rd_lock_key_slot_11)
            if sec_eng_key_sel == 3:
                efuse_data[keyslot11:keyslot12] = sec_eng_key[16:32]
                efuse_data[keyslot8:keyslot9] = sec_eng_key[0:16]
                efuse_mask_data[keyslot8:keyslot9] = mask_4bytes * 4
                efuse_mask_data[keyslot11:keyslot12] = mask_4bytes * 4
                rw_lock1 |= (1 << wr_lock_key_slot_8)
                rw_lock1 |= (1 << wr_lock_key_slot_11)
                rw_lock1 |= (1 << rd_lock_key_slot_8)
                rw_lock1 |= (1 << rd_lock_key_slot_11)
        if flash_encryp_type == 1:
            if sec_eng_key_sel == 0:
                efuse_data[keyslot8:keyslot9] = sec_eng_key[16:32]
                efuse_data[keyslot9:keyslot10] = sec_eng_key[0:16]
                efuse_mask_data[keyslot8:keyslot10] = mask_4bytes * 8
                rw_lock1 |= (1 << wr_lock_key_slot_8)
                rw_lock1 |= (1 << wr_lock_key_slot_9)
                rw_lock1 |= (1 << rd_lock_key_slot_8)
                rw_lock1 |= (1 << rd_lock_key_slot_9)
            if sec_eng_key_sel == 1:
                efuse_data[keyslot9:keyslot10] = sec_eng_key[16:32]
                efuse_data[keyslot11:keyslot12] = sec_eng_key[0:16]
                efuse_mask_data[keyslot9:keyslot10] = mask_4bytes * 4
                efuse_mask_data[keyslot11:keyslot12] = mask_4bytes * 4
                rw_lock1 |= (1 << wr_lock_key_slot_9)
                rw_lock1 |= (1 << wr_lock_key_slot_11)
                rw_lock1 |= (1 << rd_lock_key_slot_9)
                rw_lock1 |= (1 << rd_lock_key_slot_11)
            if sec_eng_key_sel == 2:
                efuse_data[keyslot11:keyslot12] = sec_eng_key[16:32]
                efuse_data[keyslot8:keyslot9] = sec_eng_key[0:16]
                efuse_mask_data[keyslot8:keyslot9] = mask_4bytes * 4
                efuse_mask_data[keyslot11:keyslot12] = mask_4bytes * 4
                rw_lock1 |= (1 << wr_lock_key_slot_8)
                rw_lock1 |= (1 << wr_lock_key_slot_11)
                rw_lock1 |= (1 << rd_lock_key_slot_8)
                rw_lock1 |= (1 << rd_lock_key_slot_11)
            if sec_eng_key_sel == 3:
                efuse_data[keyslot8:keyslot9] = sec_eng_key[16:32]
                efuse_data[keyslot11:keyslot12] = sec_eng_key[0:16]
                efuse_mask_data[keyslot8:keyslot9] = mask_4bytes * 4
                efuse_mask_data[keyslot11:keyslot12] = mask_4bytes * 4
                rw_lock1 |= (1 << wr_lock_key_slot_8)
                rw_lock1 |= (1 << wr_lock_key_slot_11)
                rw_lock1 |= (1 << rd_lock_key_slot_8)
                rw_lock1 |= (1 << rd_lock_key_slot_11)
        if flash_encryp_type == 2 or flash_encryp_type == 3:
            if sec_eng_key_sel == 0:
                efuse_data[keyslot9:keyslot10] = sec_eng_key[0:16]
                efuse_mask_data[keyslot9:keyslot10] = mask_4bytes * 4
                rw_lock1 |= (1 << wr_lock_key_slot_9)
                rw_lock1 |= (1 << rd_lock_key_slot_9)
            if sec_eng_key_sel == 1:
                efuse_data[keyslot9:keyslot10] = sec_eng_key[0:16]
                efuse_mask_data[keyslot9:keyslot10] = mask_4bytes * 4
                rw_lock1 |= (1 << wr_lock_key_slot_9)
                rw_lock1 |= (1 << rd_lock_key_slot_9)
            if sec_eng_key_sel == 2:
                efuse_data[keyslot9:keyslot10] = sec_eng_key[0:16]
                efuse_mask_data[keyslot9:keyslot10] = mask_4bytes * 4
                rw_lock1 |= (1 << wr_lock_key_slot_9)
                rw_lock1 |= (1 << rd_lock_key_slot_9)
            if sec_eng_key_sel == 3:
                efuse_data[keyslot9:keyslot10] = sec_eng_key[0:16]
                efuse_mask_data[keyslot9:keyslot10] = mask_4bytes * 4
                rw_lock1 |= (1 << wr_lock_key_slot_9)
                rw_lock1 |= (1 << rd_lock_key_slot_9)
    efuse_data[124:128] = bflb_utils.int_to_4bytearray_l(rw_lock0)
    efuse_mask_data[124:128] = bflb_utils.int_to_4bytearray_l(rw_lock0)
    efuse_data[252:256] = bflb_utils.int_to_4bytearray_l(rw_lock1)
    efuse_mask_data[252:256] = bflb_utils.int_to_4bytearray_l(rw_lock1)
    fp = open_file(cfg.get("Img_CPU1_Cfg", "efuse_file"), 'wb+')
    fp.write(efuse_data)
    fp.close()
    fp = open_file(cfg.get("Img_CPU1_Cfg", "efuse_mask_file"), 'wb+')
    fp.write(efuse_mask_data)
    fp.close()


# get sign and encrypt info
def img_create_get_sign_encrypt_info(bootheader_data):
    sign = bootheader_data[116] & 0x3
    encrypt = ((bootheader_data[116] >> 2) & 0x3)
    key_sel = ((bootheader_data[116] >> 4) & 0x3)
    return sign, encrypt, key_sel


# get hash ignore ignore
def img_create_get_hash_ignore(bootheader_data):
    return (bootheader_data[118] >> 1) & 0x1


# get crc ignore ignore
def img_create_get_crc_ignore(bootheader_data):
    return bootheader_data[118] & 0x1


# update boot header info
def img_create_update_bootheader(bootheader_data, hash, seg_cnt):
    # update segment count
    bootheader_data[120:124] = bflb_utils.int_to_4bytearray_l(seg_cnt)
    # update hash
    sign, encrypt, key_sel = img_create_get_sign_encrypt_info(bootheader_data)
    if img_create_get_hash_ignore(bootheader_data) == 1 and sign == 0:
        # do nothing
        bflb_utils.printf("Hash ignored")
    else:
        bootheader_data[132:164] = hash
    # update header crc
    if img_create_get_crc_ignore(bootheader_data) == 1:
        # do nothing
        bflb_utils.printf("Header crc ignored")
    else:
        hd_crcarray = bflb_utils.get_crc32_bytearray(bootheader_data[0:176 - 4])
        bootheader_data[176 - 4:176] = hd_crcarray
        bflb_utils.printf("Header crc: ", binascii.hexlify(hd_crcarray).decode('utf-8'))
    return bootheader_data


# update segment header according segdata
def img_create_update_segheader(segheader, segdatalen, segdatacrc):
    segheader[4:8] = segdatalen
    segheader[8:12] = segdatacrc
    return segheader


# sign image(hash code)
def img_create_sign_data(data_bytearray, privatekey_file_uecc, publickey_file):
    sk = ecdsa.SigningKey.from_pem(open_file(privatekey_file_uecc).read())
    vk = ecdsa.VerifyingKey.from_pem(open_file(publickey_file).read())
    pk_data = vk.to_string()
    bflb_utils.printf("Private key: ", binascii.hexlify(sk.to_string()))
    bflb_utils.printf("Public key: ", binascii.hexlify(pk_data))
    pk_hash = img_create_sha256_data(pk_data)
    bflb_utils.printf("Public key hash=", binascii.hexlify(pk_hash))
    signature = sk.sign(data_bytearray,
                        hashfunc=hashlib.sha256,
                        sigencode=ecdsa.util.sigencode_string)
    bflb_utils.printf("Signature=", binascii.hexlify(signature))
    # return len+signature+crc
    len_array = bflb_utils.int_to_4bytearray_l(len(signature))
    sig_field = len_array + signature
    crcarray = bflb_utils.get_crc32_bytearray(sig_field)
    return pk_data, pk_hash, sig_field + crcarray


# read one file and append crc if needed
def img_create_read_file_append_crc(file, crc):
    fp = open_file(file, 'rb')
    read_data = bytearray(fp.read())
    crcarray = bytearray(0)
    if crc:
        crcarray = bflb_utils.get_crc32_bytearray(read_data)
    fp.close()
    return read_data + crcarray


def encrypt_loader_bin_do(file, sign, encrypt, cfg):
    return False, None


def img_creat_process(img_type, flash_img, cfg):
    encrypt_blk_size = 16
    padding = bytearray(encrypt_blk_size)
    data_tohash = bytearray(0)
    if img_type == "cpu0":
        img_update_efuse_fun = img_update_efuse_cpu0
        cfg_section = "Img_CPU0_Cfg"
    elif img_type == "cpu1":
        img_update_efuse_fun = img_update_efuse_cpu1
        cfg_section = "Img_CPU1_Cfg"
    else:
        img_update_efuse_fun = img_update_efuse_sp
        cfg_section = "Img_Cfg"
    # get segdata to deal with
    segheader_file = []
    if flash_img == 0:
        for files in cfg.get(cfg_section, "segheader_file").split(" "):
            segheader_file.append(str(files))
    segdata_file = []
    for files in cfg.get(cfg_section, "segdata_file").split("|"):
        segdata_file.append(str(files))
        if flash_img == 1:
            break
    # get bootheader
    boot_header_file = cfg.get(cfg_section, "boot_header_file")
    bootheader_data = img_create_read_file_append_crc(boot_header_file, 0)
    # decide encrypt and sign
    encrypt = 0
    sign, encrypt, key_sel = img_create_get_sign_encrypt_info(bootheader_data)
    aesiv_data = bytearray(0)
    pk_data = bytearray(0)
    if sign != 0:
        bflb_utils.printf("Image need sign")
        publickey_file = cfg.get(cfg_section, "publickey_file")
        privatekey_file_uecc = cfg.get(cfg_section, "privatekey_file_uecc")
    if encrypt != 0:
        bflb_utils.printf("Image need encrypt ", encrypt)
        encrypt_key_org = bflb_utils.hexstr_to_bytearray(cfg.get(cfg_section, "aes_key_org"))
        if encrypt == 1:
            encrypt_key = encrypt_key_org[0:16]
        elif encrypt == 2:
            encrypt_key = encrypt_key_org[0:32]
        elif encrypt == 3:
            encrypt_key = encrypt_key_org[0:24]
        bflb_utils.printf("Key= ", binascii.hexlify(encrypt_key))
        encrypt_iv = bflb_utils.hexstr_to_bytearray(cfg.get(cfg_section, "aes_iv"))
        iv_crcarray = bflb_utils.get_crc32_bytearray(encrypt_iv)
        aesiv_data = encrypt_iv + iv_crcarray
        data_tohash = data_tohash + aesiv_data
    # decide seg_cnt values
    seg_cnt = len(segheader_file)
    if flash_img == 0 and seg_cnt != len(segdata_file):
        bflb_utils.printf("Segheader count and segdata count not match")
        return "FAIL", data_tohash
    data_toencrypt = bytearray(0)
    if flash_img == 0:
        i = 0
        seg_header_list = []
        seg_data_list = []
        while i < seg_cnt:
            # read seg data and calculate crcdata
            seg_data = img_create_read_file_append_crc(segdata_file[i], 0)
            padding_size = 0
            if len(seg_data) % encrypt_blk_size != 0:
                padding_size = encrypt_blk_size - len(seg_data) % encrypt_blk_size
                seg_data += padding[0:padding_size]
            segdata_crcarray = bflb_utils.get_crc32_bytearray(seg_data)
            seg_data_list.append(seg_data)

            # read seg header and replace segdata's CRC
            seg_header = img_create_read_file_append_crc(segheader_file[i], 0)
            seg_header = img_create_update_segheader(seg_header,
                                                     bflb_utils.int_to_4bytearray_l(len(seg_data)),
                                                     segdata_crcarray)
            segheader_crcarray = bflb_utils.get_crc32_bytearray(seg_header)
            seg_header = seg_header + segheader_crcarray
            seg_header_list.append(seg_header)
            i = i + 1
        # get all data to encrypt
        i = 0
        while i < seg_cnt:
            # now changed to encrypted since download tool's segdata len is from bootrom
            data_toencrypt += seg_header_list[i]
            data_toencrypt += seg_data_list[i]
            i += 1
    else:
        seg_data = img_create_read_file_append_crc(segdata_file[0], 0)
        padding_size = 0
        if len(seg_data) % encrypt_blk_size != 0:
            padding_size = encrypt_blk_size - len(seg_data) % encrypt_blk_size
            seg_data += padding[0:padding_size]
        data_toencrypt += seg_data
        seg_cnt = len(data_toencrypt)
    # do encrypt
    if encrypt != 0:
        data_toencrypt = img_create_encrypt_data(data_toencrypt, encrypt_key, encrypt_iv,
                                                 flash_img)
    # get fw data
    fw_data = bytearray(0)
    data_tohash += data_toencrypt
    fw_data = data_toencrypt
    # hash fw img
    hash = img_create_sha256_data(data_tohash)
    bflb_utils.printf("Image hash is ", binascii.hexlify(hash).decode('utf-8'))
    #update boot header and recalculate crc
    bootheader_data = img_create_update_bootheader(bootheader_data, hash, seg_cnt)
    # add signautre
    signature = bytearray(0)
    pk_hash = None
    if sign == 1:
        pk_data, pk_hash, signature = img_create_sign_data(data_tohash, privatekey_file_uecc,
                                                           publickey_file)
        pk_data = pk_data + bflb_utils.get_crc32_bytearray(pk_data)
    # write whole image
    if flash_img == 1:
        bflb_utils.printf("Write flash img")
        bootinfo_file_name = cfg.get(cfg_section, "bootinfo_file")
        fp = open_file(bootinfo_file_name, 'wb+')
        bootinfo = bootheader_data + pk_data + pk_data + signature + signature + aesiv_data
        fp.write(bootinfo)
        fp.close()
        fw_file_name = cfg.get(cfg_section, "img_file")
        fp = open_file(fw_file_name, 'wb+')
        fp.write(fw_data)
        fp.close()
        # update efuse
        if encrypt != 0:
            if encrypt == 1:
                # AES 128
                img_update_efuse_fun(cfg, sign, pk_hash, 1,
                                     encrypt_key + bytearray(32 - len(encrypt_key)), key_sel, None)
            if encrypt == 2:
                # AES 256
                img_update_efuse_fun(cfg, sign, pk_hash, 3,
                                     encrypt_key + bytearray(32 - len(encrypt_key)), key_sel, None)
            if encrypt == 3:
                # AES 192
                img_update_efuse_fun(cfg, sign, pk_hash, 2,
                                     encrypt_key + bytearray(32 - len(encrypt_key)), key_sel, None)
        else:
            img_update_efuse_fun(cfg, sign, pk_hash, encrypt, None, key_sel, None)
    else:
        bflb_utils.printf("Write if img")
        whole_img_file_name = cfg.get(cfg_section, "whole_img_file")
        fp = open_file(whole_img_file_name, 'wb+')
        img_data = bootheader_data + pk_data + pk_data + signature + signature + aesiv_data + fw_data
        fp.write(img_data)
        fp.close()
        # update efuse
        if encrypt != 0:
            img_update_efuse_fun(cfg, sign, pk_hash, 1, None, key_sel,
                                 encrypt_key + bytearray(32 - len(encrypt_key)))
        else:
            img_update_efuse_fun(cfg, sign, pk_hash, 0, None, key_sel, bytearray(32))
    return "OK", data_tohash


def bind_if_img(img_dir_path, cfg):
    # write whole image
    fp = open_file(cfg.get("Img_CPU0_Cfg", "whole_img_file"), 'rb')
    read_data = fp.read()
    fp.close()
    fp = open_file(cfg.get("Img_CPU1_Cfg", "whole_img_file"), 'rb')
    read_data = read_data + fp.read()
    fp.close()
    fp = open_file(img_dir_path + "/wholeimg_cpu0_cpu1_if.bin", 'wb+')
    fp.write(read_data)
    fp.close()


def sign_process(cfg, data_tohash, file, privatekey_file_uecc, publickey_file, signer):
    fp = open_file(file, 'rb')
    header = bytearray(fp.read())
    header = header + bytearray(0)
    fp.close()
    pk_data, pk_hash, signature_field = img_create_sign_data(data_tohash, privatekey_file_uecc,
                                                             publickey_file)
    pk_data = pk_data + bflb_utils.get_crc32_bytearray(pk_data)
    pk_pos = 244
    header[pk_pos:pk_pos + 64 + 4] = pk_data
    sign_pos = 384
    header[sign_pos:sign_pos + 4 + 64 + 4] = signature_field
    fp = open_file(file, 'wb+')
    fp.write(header)
    fp.close()

    # update public key hash
    mask_4bytes = bytearray.fromhex("FFFFFFFF")
    if signer == "cpu0":
        bflb_utils.printf("CPU0 is signer, add signature for CPU1's image")
        fp = open_file(cfg.get("Img_CPU0_Cfg", "efuse_file"), 'rb')
        efuse_data = bytearray(fp.read()) + bytearray(0)
        fp.close()
        fp = open_file(cfg.get("Img_CPU0_Cfg", "efuse_mask_file"), 'rb')
        efuse_mask_data = bytearray(fp.read()) + bytearray(0)
        fp.close()
        efuse_data[keyslot0:keyslot2] = pk_hash
        efuse_mask_data[keyslot0:keyslot2] = mask_4bytes * 8
        fp = open_file(cfg.get("Img_CPU0_Cfg", "efuse_file"), 'wb+')
        fp.write(efuse_data)
        fp.close()
        fp = open_file(cfg.get("Img_CPU0_Cfg", "efuse_mask_file"), 'wb+')
        fp.write(efuse_mask_data)
        fp.close()
    else:
        bflb_utils.printf("CPU1 is signer, add signature for CPU0's image")
        fp = open_file(cfg.get("Img_CPU1_Cfg", "efuse_file"), 'rb')
        efuse_data = bytearray(fp.read()) + bytearray(0)
        fp.close()
        fp = open_file(cfg.get("Img_CPU1_Cfg", "efuse_mask_file"), 'rb')
        efuse_mask_data = bytearray(fp.read()) + bytearray(0)
        fp.close()
        efuse_data[keyslot5:keyslot5 + 16] = pk_hash[0:16]
        efuse_mask_data[keyslot5:keyslot5 + 16] = mask_4bytes * 4
        efuse_data[keyslot6:keyslot7] = pk_hash[16:32]
        efuse_mask_data[keyslot6:keyslot7] = mask_4bytes * 4
        fp = open_file(cfg.get("Img_CPU1_Cfg", "efuse_file"), 'wb+')
        fp.write(efuse_data)
        fp.close()
        fp = open_file(cfg.get("Img_CPU1_Cfg", "efuse_mask_file"), 'wb+')
        fp.write(efuse_mask_data)
        fp.close()


def img_create_do(args, img_dir_path=None, config_file=None):
    bflb_utils.printf("========= image create =========")
    bflb_utils.printf("Image create path: ", img_dir_path)
    if config_file is None:
        config_file = img_dir_path + "/img_create_cfg.ini"
    bflb_utils.printf("Config file: ", config_file)
    cfg = BFConfigParser()
    cfg.read(config_file)
    cpu_type = ""
    signer = "none"
    ret0 = "OK"
    ret1 = "OK"
    data_tohash_cpu0 = bytearray(0)
    data_tohash_cpu1 = bytearray(0)
    try:
        if args.image:
            img_type = args.image
        if args.cpu:
            cpu_type = args.cpu
        if args.signer:
            signer = args.signer
    except Exception as e:
        # bflb_utils.printf(help information and exit:)
        bflb_utils.printf(e)  # will something like "option -a not recognized")
    if img_type == "media":
        flash_img = 1
    else:
        flash_img = 0

    # deal image creation
    if cpu_type == "cpu0":
        ret0, data_tohash_cpu0 = img_creat_process("cpu0", flash_img, cfg)
    elif cpu_type == "cpu1":
        ret1, data_tohash_cpu1 = img_creat_process("cpu1", flash_img, cfg)
    elif cpu_type == "all":
        ret0, data_tohash_cpu0 = img_creat_process("cpu0", flash_img, cfg)
        ret1, data_tohash_cpu1 = img_creat_process("cpu1", flash_img, cfg)
    else:
        img_creat_process("", flash_img, cfg)
        return  # shoud return here for sp
    if ret0 == "OK" and ret1 == "OK":
        if signer == "cpu0":
            bflb_utils.printf("cpu0 sign cpu1 ")
            if flash_img == 1:
                sign_process(cfg, data_tohash_cpu1, cfg.get("Img_CPU1_Cfg", "bootinfo_file"),
                             cfg.get("Img_CPU0_Cfg", "privatekey_file_uecc"),
                             cfg.get("Img_CPU0_Cfg", "publickey_file"), "cpu0")
            else:
                sign_process(cfg, data_tohash_cpu1, cfg.get("Img_CPU1_Cfg", "whole_img_file"),
                             cfg.get("Img_CPU0_Cfg", "privatekey_file_uecc"),
                             cfg.get("Img_CPU0_Cfg", "publickey_file"), "cpu0")
        elif signer == "cpu1":
            bflb_utils.printf("cpu1 sign cpu0 ")
            if flash_img == 1:
                sign_process(cfg, data_tohash_cpu0, cfg.get("Img_CPU0_Cfg", "bootinfo_file"),
                             cfg.get("Img_CPU1_Cfg", "privatekey_file_uecc"),
                             cfg.get("Img_CPU1_Cfg", "publickey_file"), "cpu1")
            else:
                sign_process(cfg, data_tohash_cpu0, cfg.get("Img_CPU0_Cfg", "whole_img_file"),
                             cfg.get("Img_CPU1_Cfg", "privatekey_file_uecc"),
                             cfg.get("Img_CPU1_Cfg", "publickey_file"), "cpu1")
    else:
        bflb_utils.printf("Fail to create images!")
        return

    # if user select all cpu, bind sdio img
    if cpu_type == "all" and flash_img == 0:
        bind_if_img(img_dir_path, cfg)


def create_sp_media_image(config, cpu_type=None):
    bflb_utils.printf("========= sp image create =========")
    cfg = BFConfigParser()
    cfg.read(config)
    img_creat_process("cpu0", 1, cfg)



# -*- coding:utf-8 -*-

import os
import re
from struct import unpack_from, unpack, pack
from string import printable

########################################################################################################################
# misc.py
########################################################################################################################


def is_string(data):
    """ Check property string validity """
    if not len(data):
        return None
    if data[-1] != 0:
        return None
    pos = 0
    while pos < len(data):
        posi = pos
        while pos < len(data) and \
              data[pos] != 0 and \
              data[pos] in printable.encode() and \
              data[pos] not in (ord('\r'), ord('\n')):
            pos += 1
        if data[pos] != 0 or pos == posi:
            return None
        pos += 1
    return True


def extract_string(data, offset=0):
    """ Extract string """
    str_end = offset
    while data[str_end] != 0:
        str_end += 1
    return data[offset:str_end].decode("ascii")


def line_offset(tabsize, offset, string):
    offset = " " * (tabsize * offset)
    return offset + string


def get_version_info(text):
    ret = dict()
    for line in text.split('\n'):
        line = line.rstrip('\0')
        if line and line.startswith('/ {'):
            break
        if line and line.startswith('//'):
            line = line.replace('//', '').replace(':', '')
            line = line.split()
            if line[0] in ('version', 'last_comp_version', 'boot_cpuid_phys'):
                ret[line[0]] = int(line[1], 0)
    return ret


def strip_comments(text):
    text = re.sub('//.*?(\r\n?|\n)|/\*.*?\*/', '\n', text, flags=re.S)
    return text


def split_to_lines(text):
    lines = []
    mline = str()
    for line in text.split('\n'):
        line = line.replace('\t', ' ')
        line = line.rstrip('\0')
        line = line.rstrip(' ')
        line = line.lstrip(' ')
        if not line or line.startswith('/dts-'):
            continue
        if line.endswith('{') or line.endswith(';'):
            line = line.replace(';', '')
            lines.append(mline + line)
            mline = str()
        else:
            mline += line

    return lines


########################################################################################################################
# Binary Blob Constants
########################################################################################################################

DTB_BEGIN_NODE = 0x1
DTB_END_NODE = 0x2
DTB_PROP = 0x3
DTB_NOP = 0x4
DTB_END = 0x9

########################################################################################################################
# Header Class
########################################################################################################################


class Header:

    MIN_SIZE = 4 * 7
    MAX_SIZE = 4 * 10

    MAX_VERSION = 17

    MAGIC_NUMBER = 0xD00DFEED

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        if value > self.MAX_VERSION:
            raise ValueError(f"Invalid Version {value}, use: 0 - 17 !")
        # update size and padding
        self._size = self.MIN_SIZE
        if value >= 2:
            self._size += 4
        if value >= 3:
            self._size += 4
        if value >= 17:
            self._size += 4
        self._padding = 8 - (self._size % 8) if self._size % 8 != 0 else 0
        self._version = value
        self.last_comp_version = value - 1

    @property
    def size(self):
        return self._size + self._padding

    @property
    def padding(self):
        return self._padding

    def __init__(self):
        # private variables
        self._version = None
        self._size = 0
        self._padding = 0
        # public variables
        self.total_size = 0
        self.off_dt_struct = 0
        self.off_dt_strings = 0
        self.off_mem_rsvmap = 0
        self.last_comp_version = 0
        # version depend variables
        self.boot_cpuid_phys = 0
        self.size_dt_strings = None
        self.size_dt_struct = None

    def __str__(self):
        return '<FDT-v{}, size: {}>'.format(self.version, self.size)

    def info(self):
        nfo = 'FDT Header:'
        nfo += '- Version: {}'.format(self.version)
        nfo += '- Size:    {}'.format(self.size)
        return nfo

    def export(self) -> bytes:
        """

        :return:
        """
        if self.version is None:
            raise Exception("Header Version must be specified !")

        blob = pack('>7I', self.MAGIC_NUMBER, self.total_size, self.off_dt_struct,
                    self.off_dt_strings, self.off_mem_rsvmap, self.version, self.last_comp_version)
        if self.version >= 2:
            blob += pack('>I', self.boot_cpuid_phys)
        if self.version >= 3:
            blob += pack('>I', self.size_dt_strings)
        if self.version >= 17:
            blob += pack('>I', self.size_dt_struct)
        if self.padding:
            blob += bytes([0] * self.padding)

        return blob

    @classmethod
    def parse(cls, data: bytes, offset: int = 0):
        """

        :param data:
        :param offset:
        """
        if len(data) < (offset + cls.MIN_SIZE):
            raise ValueError('Data size too small !')

        header = cls()
        (magic_number, header.total_size, header.off_dt_struct, header.off_dt_strings,
         header.off_mem_rsvmap, header.version,
         header.last_comp_version) = unpack_from('>7I', data, offset)
        offset += cls.MIN_SIZE

        if magic_number != cls.MAGIC_NUMBER:
            raise Exception('Invalid Magic Number')
        if header.last_comp_version > cls.MAX_VERSION - 1:
            raise Exception(f'Invalid last compatible Version {header.last_comp_version}')

        if header.version >= 2:
            header.boot_cpuid_phys = unpack_from('>I', data, offset)[0]
            offset += 4

        if header.version >= 3:
            header.size_dt_strings = unpack_from('>I', data, offset)[0]
            offset += 4

        if header.version >= 17:
            header.size_dt_struct = unpack_from('>I', data, offset)[0]
            offset += 4

        return header


# items.py
########################################################################################################################
# Helper methods
########################################################################################################################


def new_property(name: str, raw_value: bytes) -> object:
    """
    Instantiate property with raw value type

    :param name: Property name
    :param raw_value: Property raw data
    """
    if is_string(raw_value):
        obj = PropStrings(name)
        # Extract strings from raw value
        for st in raw_value.decode('ascii').split('\0'):
            if st:
                obj.append(st)
        return obj

    elif len(raw_value) and len(raw_value) % 4 == 0:
        obj = PropWords(name)
        # Extract words from raw value
        for i in range(0, len(raw_value), 4):
            obj.append(unpack(">I", raw_value[i:i + 4])[0])
        return obj

    elif len(raw_value):
        return PropBytes(name, data=raw_value)

    else:
        return Property(name)


########################################################################################################################
# Base Class
########################################################################################################################


class BaseItem:

    @property
    def name(self):
        return self._name

    @property
    def parent(self):
        return self._parent

    @property
    def path(self):
        node = self._parent
        path = ""
        while node:
            if node.name == '/':
                break
            path = '/' + node.name + path
            node = node.parent
        return path if path else '/'

    def __init__(self, name: str):
        """ 
        BaseItem constructor
        
        :param name: Item name
        """
        assert isinstance(name, str)
        assert all(c in printable for c in name), "The value must contain just printable chars !"
        self._name = name
        self._parent = None

    def __str__(self):
        """ String representation """
        return f"{self.name}"

    def set_name(self, value: str):
        """ 
        Set item name
        
        :param value: The name in string format
        """
        assert isinstance(value, str)
        assert all(c in printable for c in value), "The value must contain just printable chars !"
        self._name = value

    def set_parent(self, value):
        """ 
        Set item parent

        :param value: The parent node 
        """
        assert isinstance(value, Node)
        self._parent = value

    def to_dts(self, tabsize: int = 4, depth: int = 0):
        raise NotImplementedError()

    def to_dtb(self, strings: str, pos: int = 0, version: int = Header.MAX_VERSION):
        raise NotImplementedError()


########################################################################################################################
# Property Classes
########################################################################################################################


class Property(BaseItem):

    def __getitem__(self, value):
        """ Returns No Items """
        return None

    def __eq__(self, obj):
        """ Check Property object equality """
        return isinstance(obj, Property) and self.name == obj.name

    def copy(self):
        """ Get object copy """
        return Property(self.name)

    def to_dts(self, tabsize: int = 4, depth: int = 0):
        """
        Get string representation

        :param tabsize: Tabulator size in count of spaces
        :param depth: Start depth for line
        """
        return line_offset(tabsize, depth, '{};\n'.format(self.name))

    def to_dtb(self, strings: str, pos: int = 0, version: int = Header.MAX_VERSION):
        """
        Get binary blob representation

        :param strings:
        :param pos:
        :param version:
        """
        strpos = strings.find(self.name + '\0')
        if strpos < 0:
            strpos = len(strings)
            strings += self.name + '\0'
        pos += 12
        return pack('>III', DTB_PROP, 0, strpos), strings, pos


class PropStrings(Property):
    """Property with strings as value"""

    @property
    def value(self):
        return self.data[0] if self.data else None

    def __init__(self, name: str, *args):
        """ 
        PropStrings constructor
        
        :param name: Property name
        :param args: str1, str2, ...
        """
        super().__init__(name)
        self.data = []
        for arg in args:
            self.append(arg)

    def __str__(self):
        """ String representation """
        return f"{self.name} = {self.data}"

    def __len__(self):
        """ Get strings count """
        return len(self.data)

    def __getitem__(self, index):
        """ Get string by index """
        return self.data[index]

    def __eq__(self, obj):
        """ Check PropStrings object equality """
        if not isinstance(obj, PropStrings) or self.name != obj.name or len(self) != len(obj):
            return False
        for index in range(len(self)):
            if self.data[index] != obj[index]:
                return False
        return True

    def copy(self):
        """ Get object copy """
        return PropStrings(self.name, *self.data)

    def append(self, value: str):
        assert isinstance(value, str)
        assert len(value) > 0, "Invalid strings value"
        assert all(
            c in printable or c in ('\r', '\n') for c in value), "Invalid chars in strings value"
        self.data.append(value)

    def pop(self, index: int):
        assert 0 <= index < len(self.data), "Index out of range"
        return self.data.pop(index)

    def clear(self):
        self.data.clear()

    def to_dts(self, tabsize: int = 4, depth: int = 0):
        """
        Get string representation

        :param tabsize: Tabulator size in count of spaces
        :param depth: Start depth for line
        """
        result = line_offset(tabsize, depth, self.name)
        result += ' = "'
        result += '", "'.join(self.data)
        result += '";\n'
        return result

    def to_dtb(self, strings: str, pos: int = 0, version: int = Header.MAX_VERSION):
        """
        Get blob representation

        :param strings:
        :param pos:
        :param version:
        """
        blob = pack('')
        for chars in self.data:
            blob += chars.encode('ascii') + pack('b', 0)
        blob_len = len(blob)
        if version < 16 and (pos + 12) % 8 != 0:
            blob = pack('b', 0) * (8 - ((pos + 12) % 8)) + blob
        if blob_len % 4:
            blob += pack('b', 0) * (4 - (blob_len % 4))
        strpos = strings.find(self.name + '\0')
        if strpos < 0:
            strpos = len(strings)
            strings += self.name + '\0'
        blob = pack('>III', DTB_PROP, blob_len, strpos) + blob
        pos += len(blob)
        return blob, strings, pos


class PropWords(Property):
    """Property with words as value"""

    @property
    def value(self):
        return self.data[0] if self.data else None

    def __init__(self, name, *args):
        """
        PropWords constructor

        :param name: Property name
        :param args: word1, word2, ...
        """
        super().__init__(name)
        self.data = []
        self.word_size = 32
        for val in args:
            self.append(val)

    def __str__(self):
        """ String representation """
        return f"{self.name} = {self.data}"

    def __getitem__(self, index):
        """ Get word by index """
        return self.data[index]

    def __len__(self):
        """ Get words count """
        return len(self.data)

    def __eq__(self, prop):
        """ Check PropWords object equality  """
        if not isinstance(prop, PropWords):
            return False
        if self.name != prop.name:
            return False
        if len(self) != len(prop):
            return False
        for index in range(len(self)):
            if self.data[index] != prop[index]:
                return False
        return True

    def copy(self):
        return PropWords(self.name, *self.data)

    def append(self, value):
        assert isinstance(value, int), "Invalid object type"
        assert 0 <= value < 2**self.word_size, "Invalid word value {}, use <0x0 - 0x{:X}>".format(
            value, 2**self.word_size - 1)
        self.data.append(value)

    def pop(self, index):
        assert 0 <= index < len(self.data), "Index out of range"
        return self.data.pop(index)

    def clear(self):
        self.data.clear()

    def to_dts(self, tabsize: int = 4, depth: int = 0):
        """
        Get string representation

        :param tabsize: Tabulator size in count of spaces
        :param depth: Start depth for line
        """
        result = line_offset(tabsize, depth, self.name)
        result += ' = <'
        result += ' '.join(["0x{:X}".format(word) for word in self.data])
        result += ">;\n"
        return result

    def to_dtb(self, strings: str, pos: int = 0, version: int = Header.MAX_VERSION):
        """
        Get blob representation

        :param strings:
        :param pos:
        :param version:
        """
        strpos = strings.find(self.name + '\0')
        if strpos < 0:
            strpos = len(strings)
            strings += self.name + '\0'
        blob = pack('>III', DTB_PROP, len(self.data) * 4, strpos)
        for word in self.data:
            blob += pack('>I', word)
        pos += len(blob)
        return blob, strings, pos


class PropBytes(Property):
    """Property with bytes as value"""

    def __init__(self, name, data=None):
        """ 
        PropBytes constructor
        
        :param name: Property name
        :param data: Data as list, bytes or bytearray
        """
        super().__init__(name)
        self.data = bytearray()
        if data:
            assert isinstance(data, (list, bytes, bytearray))
            self.data = bytearray(data)

    def __str__(self):
        """ String representation """
        return f"{self.name} = {self.data}"

    def __getitem__(self, index):
        """Get byte by index """
        return self.data[index]

    def __len__(self):
        """ Get bytes count """
        return len(self.data)

    def __eq__(self, prop):
        """ Check PropBytes object equality  """
        if not isinstance(prop, PropBytes):
            return False
        if self.name != prop.name:
            return False
        if len(self) != len(prop):
            return False
        for index in range(len(self)):
            if self.data[index] != prop[index]:
                return False
        return True

    def copy(self):
        """ Create a copy of object """
        return PropBytes(self.name, self.data)

    def append(self, value):
        assert isinstance(value, int), "Invalid object type"
        assert 0 <= value <= 0xFF, "Invalid byte value {}, use <0 - 255>".format(value)
        self.data.append(value)

    def pop(self, index):
        assert 0 <= index < len(self.data), "Index out of range"
        return self.data.pop(index)

    def clear(self):
        self.data = bytearray()

    def to_dts(self, tabsize: int = 4, depth: int = 0):
        """
        Get string representation

        :param tabsize: Tabulator size in count of spaces
        :param depth: Start depth for line
        """
        result = line_offset(tabsize, depth, self.name)
        result += ' = ['
        result += ' '.join(["{:02X}".format(byte) for byte in self.data])
        result += '];\n'
        return result

    def to_dtb(self, strings: str, pos: int = 0, version: int = Header.MAX_VERSION):
        """
        Get blob representation

        :param strings:
        :param pos:
        :param version:
        """
        strpos = strings.find(self.name + '\0')
        if strpos < 0:
            strpos = len(strings)
            strings += self.name + '\0'
        blob = pack('>III', DTB_PROP, len(self.data), strpos)
        blob += bytes(self.data)
        if len(blob) % 4:
            blob += bytes([0] * (4 - (len(blob) % 4)))
        pos += len(blob)
        return blob, strings, pos


class PropIncBin(PropBytes):
    """Property with bytes as value"""

    def __init__(self, name, data=None, file_name=None, rpath=None):
        """
        PropIncBin constructor

        :param name: Property name
        :param data: Data as list, bytes or bytearray
        :param file_name: File name
        :param rpath: Relative path
        """
        super().__init__(name, data)
        self.file_name = file_name
        self.relative_path = rpath

    def __eq__(self, prop):
        """ Check PropIncBin object equality  """
        if not isinstance(prop, PropIncBin):
            return False
        if self.name != prop.name:
            return False
        if self.file_name != prop.file_name:
            return False
        if self.relative_path != prop.relative_path:
            return False
        if self.data != prop.data:
            return False
        return True

    def copy(self):
        """ Create a copy of object """
        return PropIncBin(self.name, self.data, self.file_name, self.relative_path)

    def to_dts(self, tabsize: int = 4, depth: int = 0):
        """
        Get string representation

        :param tabsize: Tabulator size in count of spaces
        :param depth: Start depth for line
        """
        file_path = self.file_name
        if self.relative_path is not None:
            file_path = "{}/{}".format(self.relative_path, self.file_name)
        result = line_offset(tabsize, depth, self.name)
        result += " = /incbin/(\"{}\");\n".format(file_path)
        return result


########################################################################################################################
# Node Class
########################################################################################################################


class Node(BaseItem):
    """Node representation"""

    @property
    def props(self):
        return self._props

    @property
    def nodes(self):
        return self._nodes

    @property
    def empty(self):
        return False if self.nodes or self.props else True

    def __init__(self, name, *args):
        """ 
        Node constructor
        
        :param name: Node name
        :param args: List of properties and subnodes
        """
        super().__init__(name)
        self._props = []
        self._nodes = []
        for item in args:
            self.append(item)

    def __str__(self):
        """ String representation """
        return "< {}: {} props, {} nodes >".format(self.name, len(self.props), len(self.nodes))

    def __eq__(self, node):
        """ Check node equality """
        if not isinstance(node, Node):
            return False
        if self.name != node.name or \
           len(self.props) != len(node.props) or \
           len(self.nodes) != len(node.nodes):
            return False
        for p in self.props:
            if p not in node.props:
                return False
        for n in self.nodes:
            if n not in node.nodes:
                return False
        return True

    def copy(self):
        """ Create a copy of Node object """
        node = Node(self.name)
        for p in self.props:
            node.append(p.copy())
        for n in self.nodes:
            node.append(n.copy())
        return node

    def get_property(self, name):
        """ 
        Get property object by its name
        
        :param name: Property name
        """
        for p in self.props:
            if p.name == name:
                return p
        return None

    def set_property(self, name, value):
        """
        Set property

        :param name: Property name
        :param value: Property value
        """
        if value is None:
            new_prop = Property(name)
        elif isinstance(value, int):
            new_prop = PropWords(name, value)
        elif isinstance(value, str):
            new_prop = PropStrings(name, value)
        elif isinstance(value, list) and isinstance(value[0], int):
            new_prop = PropWords(name, *value)
        elif isinstance(value, list) and isinstance(value[0], str):
            new_prop = PropStrings(name, *value)
        elif isinstance(value, (bytes, bytearray)):
            new_prop = PropBytes(name, data=value)
        else:
            raise TypeError('Value type not supported')
        new_prop.set_parent(self)
        old_prop = self.get_property(name)
        if old_prop is None:
            self.props.append(new_prop)
        else:
            index = self.props.index(old_prop)
            self.props[index] = new_prop

    def get_subnode(self, name: str):
        """ 
        Get subnode object by name

        :param name: Subnode name
        """
        for n in self.nodes:
            if n.name == name:
                return n
        return None

    def exist_property(self, name: str) -> bool:
        """ 
        Check if property exist and return True if exist else False
        
        :param name: Property name
        """
        return False if self.get_property(name) is None else True

    def exist_subnode(self, name: str) -> bool:
        """ 
        Check if subnode exist and return True if exist else False
        
        :param name: Subnode name
        """
        return False if self.get_subnode(name) is None else True

    def remove_property(self, name: str):
        """ 
        Remove property object by its name.
        
        :param name: Property name
        """
        item = self.get_property(name)
        if item is not None:
            self.props.remove(item)

    def remove_subnode(self, name: str):
        """ 
        Remove subnode object by its name.
        
        :param name: Subnode name
        """
        item = self.get_subnode(name)
        if item is not None:
            self.nodes.remove(item)

    def append(self, item):
        """ 
        Append node or property
        
        :param item: The node or property object
        """
        assert isinstance(item,
                          (Node, Property)), "Invalid object type, use \"Node\" or \"Property\""

        if isinstance(item, Property):
            if self.get_property(item.name) is not None:
                raise Exception("{}: \"{}\" property already exists".format(self, item.name))
            item.set_parent(self)
            self.props.append(item)

        else:
            if self.get_subnode(item.name) is not None:
                raise Exception("{}: \"{}\" node already exists".format(self, item.name))
            if item is self:
                raise Exception("{}: append the same node {}".format(self, item.name))
            item.set_parent(self)
            self.nodes.append(item)

    def merge(self, node_obj, replace: bool = True):
        """ 
        Merge two nodes
        
        :param node_obj: Node object
        :param replace: If True, replace current properties with the given properties
        """
        assert isinstance(node_obj, Node), "Invalid object type"

        def get_property_index(name):
            for i, p in enumerate(self.props):
                if p.name == name:
                    return i
            return None

        def get_subnode_index(name):
            for i, n in enumerate(self.nodes):
                if n.name == name:
                    return i
            return None

        for prop in node_obj.props:
            index = get_property_index(prop.name)
            if index is None:
                self.append(prop.copy())
            elif prop in self._props:
                continue
            elif replace:
                new_prop = prop.copy()
                new_prop.set_parent(self)
                self._props[index] = new_prop
            else:
                pass

        for sub_node in node_obj.nodes:
            index = get_subnode_index(sub_node.name)
            if index is None:
                self.append(sub_node.copy())
            elif sub_node in self._nodes:
                continue
            else:
                self._nodes[index].merge(sub_node, replace)

    def to_dts(self, tabsize: int = 4, depth: int = 0) -> str:
        """ 
        Get string representation of NODE object
        
        :param tabsize: Tabulator size in count of spaces
        :param depth: Start depth for line
        """
        dts = line_offset(tabsize, depth, self.name + ' {\n')
        dts += ''.join(prop.to_dts(tabsize, depth + 1) for prop in self._props)
        dts += ''.join(node.to_dts(tabsize, depth + 1) for node in self._nodes)
        dts += line_offset(tabsize, depth, "};\n")
        return dts

    def to_dtb(self, strings: str, pos: int = 0, version: int = Header.MAX_VERSION) -> tuple:
        """ 
        Get NODE in binary blob representation
        
        :param strings: 
        :param pos:
        :param version:
        """
        if self.name == '/':
            blob = pack('>II', DTB_BEGIN_NODE, 0)
        else:
            blob = pack('>I', DTB_BEGIN_NODE)
            blob += self.name.encode('ascii') + b'\0'
        if len(blob) % 4:
            blob += pack('b', 0) * (4 - (len(blob) % 4))
        pos += len(blob)
        for prop in self._props:
            (data, strings, pos) = prop.to_dtb(strings, pos, version)
            blob += data
        for node in self._nodes:
            (data, strings, pos) = node.to_dtb(strings, pos, version)
            blob += data
        pos += 4
        blob += pack('>I', DTB_END_NODE)
        return blob, strings, pos


class ItemType:
    NODE = 0
    PROP = 1
    BOTH = 3


class FDT:
    """ Flattened Device Tree Class """

    @property
    def empty(self):
        return self.root.empty

    def __init__(self, header=None):
        """
        FDT class constructor

        :param header:
        """
        self.entries = []
        self.header = Header() if header is None else header
        self.root = Node('/')

    def __str__(self):
        """ String representation """
        return self.info()

    def info(self):
        """ Return object info in human readable format """
        msg = "FDT Content:\n"
        for path, nodes, props in self.walk():
            msg += "{} [{}N, {}P]\n".format(path, len(nodes), len(props))
        return msg

    def get_node(self, path: str, create: bool = False) -> Node:
        """ 
        Get node object from specified path
        
        :param path: Path as string
        :param create: If True, not existing nodes will be created
        """
        assert isinstance(path, str), "Node path must be a string type !"

        node = self.root
        path = path.lstrip('/')
        if path:
            names = path.split('/')
            for name in names:
                item = node.get_subnode(name)
                if item is None:
                    if create:
                        item = Node(name)
                        node.append(item)
                    else:
                        raise ValueError("Path \"{}\" doesn't exists".format(path))
                node = item

        return node

    def get_property(self, name: str, path: str = '') -> Property:
        """ 
        Get property object by name from specified path
        
        :param name: Property name
        :param path: Path to sub-node
        """
        return self.get_node(path).get_property(name)

    def set_property(self, name: str, value, path: str = '', create: bool = True):
        """
        Set property object by name
        
        :param name: Property name
        :param value: Property value
        :param path: Path to subnode
        :param create: If True, not existing nodes will be created
        """
        self.get_node(path, create).set_property(name, value)

    def exist_node(self, path: str) -> bool:
        """ 
        Check if <path>/node exist and return True
        
        :param path: path/node name
        :return True if <path>/node exist else False
        """
        try:
            self.get_node(path)
        except ValueError:
            return False
        else:
            return True

    def exist_property(self, name: str, path: str = '') -> bool:
        """ 
        Check if property exist
        
        :param name: Property name
        :param path: The path
        """
        return self.get_node(path).exist_property(name) if self.exist_node(path) else False

    def remove_node(self, name: str, path: str = ''):
        """ 
        Remove node obj by path/name. Raises ValueError if path/name doesn't exist
        
        :param name: Node name
        :param path: Path to sub-node
        """
        self.get_node(path).remove_subnode(name)

    def remove_property(self, name: str, path: str = ''):
        """ 
        Remove property obj by name. Raises ValueError if path/name doesn't exist
        
        :param name: Property name
        :param path: Path to subnode
        """
        self.get_node(path).remove_property(name)

    def add_item(self, obj, path: str = '', create: bool = True):
        """ 
        Add sub-node or property at specified path. Raises ValueError if path doesn't exist
        
        :param obj: The node or property object
        :param path: The path to subnode
        :param create: If True, not existing nodes will be created
        """
        self.get_node(path, create).append(obj)

    def search(self, name: str, itype: int = ItemType.BOTH, path: str = '') -> list:
        """ 
        Search properties and/or nodes with specified name. Return list of founded items
        
        :param name: Property or Node name
        :param itype: Item type - NODE, PROP or BOTH
        :param path: Path to root node
        """
        assert isinstance(name, str), "Property name must be a string type !"

        node = self.get_node(path)
        nodes = []
        items = []
        while True:
            nodes += node.nodes
            if itype == ItemType.NODE or itype == ItemType.BOTH:
                if node.name == name:
                    items.append(node)
            if itype == ItemType.PROP or itype == ItemType.BOTH:
                for p in node.props:
                    if p.name == name:
                        items.append(p)
            if not nodes:
                break
            node = nodes.pop()

        return items

    def walk(self, path: str = '', relative: bool = False) -> list:
        """ 
        Walk trough nodes and return relative/absolute path with list of sub-nodes and properties
        
        :param path: The path to root node
        :param relative: True for relative or False for absolute return path
        """
        all_nodes = []

        node = self.get_node(path)
        while True:
            all_nodes += node.nodes
            current_path = f"{node.path}/{node.name}"
            current_path = current_path.replace('///', '/')
            current_path = current_path.replace('//', '/')
            if path and relative:
                current_path = current_path.replace(path, '').lstrip('/')
            yield (current_path, node.nodes, node.props)
            if not all_nodes:
                break
            node = all_nodes.pop()

    def merge(self, fdt_obj, replace: bool = True):
        """
        Merge external FDT object into this object.
        
        :param fdt_obj: The FDT object which will be merged into this
        :param replace: True for replace existing items or False for keep old items
        """
        assert isinstance(fdt_obj, FDT)
        if self.header.version is None:
            self.header = fdt_obj.header
        else:
            if fdt_obj.header.version is not None and \
               fdt_obj.header.version > self.header.version:
                self.header.version = fdt_obj.header.version
        if fdt_obj.entries:
            for in_entry in fdt_obj.entries:
                exist = False
                for index in range(len(self.entries)):
                    if self.entries[index]['address'] == in_entry['address']:
                        self.entries[index]['address'] = in_entry['size']
                        exist = True
                        break
                if not exist:
                    self.entries.append(in_entry)

        self.root.merge(fdt_obj.get_node('/'), replace)

    def update_phandles(self):
        all_nodes = []
        phandle_value = 0
        no_phandle_nodes = []

        node = self.root
        all_nodes += self.root.nodes
        while all_nodes:
            props = (node.get_property('phandle'), node.get_property('linux,phandle'))
            value = None
            for i, p in enumerate(props):
                if isinstance(p, PropWords) and isinstance(p.value, int):
                    value = None if i == 1 and p.value != value else p.value
            if value is None:
                no_phandle_nodes.append(node)
            elif phandle_value < value:
                phandle_value = value
            # ...
            node = all_nodes.pop()
            all_nodes += node.nodes

        if phandle_value > 0:
            phandle_value += 1

        for node in no_phandle_nodes:
            node.set_property('linux,phandle', phandle_value)
            node.set_property('phandle', phandle_value)
            phandle_value += 1

    def to_dts(self, tabsize: int = 4) -> str:
        """
        Store FDT Object into string format (DTS)

        :param tabsize:
        """
        result = "/dts-v1/;\n"
        if self.header.version is not None:
            result += "// version: {}\n".format(self.header.version)
            result += "// last_comp_version: {}\n".format(self.header.last_comp_version)
            if self.header.version >= 2:
                result += "// boot_cpuid_phys: 0x{:X}\n".format(self.header.boot_cpuid_phys)
        result += '\n'
        if self.entries:
            for entry in self.entries:
                result += "/memreserve/ "
                result += "{:#x} ".format(entry['address']) if entry['address'] else "0 "
                result += "{:#x}".format(entry['size']) if entry['size'] else "0"
                result += ";\n"
        if self.root is not None:
            result += self.root.to_dts(tabsize)
        return result

    def to_dtb(self,
               version: int = None,
               last_comp_version: int = None,
               boot_cpuid_phys: int = None) -> bytes:
        """
        Export FDT Object into Binary Blob format (DTB)

        :param version:
        :param last_comp_version:
        :param boot_cpuid_phys:
        """
        if self.root is None:
            return b''

        from struct import pack

        if version is not None:
            self.header.version = version
        if last_comp_version is not None:
            self.header.last_comp_version = last_comp_version
        if boot_cpuid_phys is not None:
            self.header.boot_cpuid_phys = boot_cpuid_phys
        if self.header.version is None:
            raise Exception("DTB Version must be specified !")

        blob_entries = bytes()
        if self.entries:
            for entry in self.entries:
                blob_entries += pack('>QQ', entry['address'], entry['size'])
        blob_entries += pack('>QQ', 0, 0)
        blob_data_start = self.header.size + len(blob_entries)
        (blob_data, blob_strings, data_pos) = self.root.to_dtb('', blob_data_start,
                                                               self.header.version)
        blob_data += pack('>I', DTB_END)
        self.header.size_dt_strings = len(blob_strings)
        self.header.size_dt_struct = len(blob_data)
        self.header.off_mem_rsvmap = self.header.size
        self.header.off_dt_struct = blob_data_start
        self.header.off_dt_strings = blob_data_start + len(blob_data)
        self.header.total_size = blob_data_start + len(blob_data) + len(blob_strings)
        blob_header = self.header.export()
        return blob_header + blob_entries + blob_data + blob_strings.encode('ascii')


def parse_dts(text: str, root_dir: str = '') -> FDT:
    """
    Parse DTS text file and create FDT Object

    :param text:
    :param root_dir: 
    """
    ver = get_version_info(text)
    text = strip_comments(text)
    dts_lines = split_to_lines(text)
    fdt_obj = FDT()
    if 'version' in ver:
        fdt_obj.header.version = ver['version']
    if 'last_comp_version' in ver:
        fdt_obj.header.last_comp_version = ver['last_comp_version']
    if 'boot_cpuid_phys' in ver:
        fdt_obj.header.boot_cpuid_phys = ver['boot_cpuid_phys']
    # parse entries
    fdt_obj.entries = []
    for line in dts_lines:
        if line.endswith('{'):
            break
        if line.startswith('/memreserve/'):
            line = line.strip(';')
            line = line.split()
            if len(line) != 3:
                raise Exception()
            fdt_obj.entries.append({'address': int(line[1], 0), 'size': int(line[2], 0)})
    # parse nodes
    curnode = None
    fdt_obj.root = None
    for line in dts_lines:
        if line.endswith('{'):
            # start node
            node_name = line.split()[0]
            new_node = Node(node_name)
            if fdt_obj.root is None:
                fdt_obj.root = new_node
            if curnode is not None:
                curnode.append(new_node)
            curnode = new_node
        elif line.endswith('}'):
            # end node
            if curnode is not None:
                curnode = curnode.parent
        else:
            # properties
            if line.find('=') == -1:
                prop_name = line
                prop_obj = Property(prop_name)
            else:
                line = line.split('=', maxsplit=1)
                prop_name = line[0].rstrip(' ')
                prop_value = line[1].lstrip(' ')
                if prop_value.startswith('<'):
                    prop_obj = PropWords(prop_name)
                    prop_value = prop_value.replace('<', '').replace('>', '')
                    for prop in prop_value.split():
                        if prop.startswith('0x'):
                            prop_obj.append(int(prop, 16))
                        elif prop.startswith('0b'):
                            prop_obj.append(int(prop, 2))
                        elif prop.startswith('0'):
                            prop_obj.append(int(prop, 8))
                        else:
                            prop_obj.append(int(prop))
                elif prop_value.startswith('['):
                    prop_obj = PropBytes(prop_name)
                    prop_value = prop_value.replace('[', '').replace(']', '')
                    for prop in prop_value.split():
                        prop_obj.append(int(prop, 16))
                elif prop_value.startswith('/incbin/'):
                    prop_value = prop_value.replace('/incbin/("', '').replace('")', '')
                    prop_value = prop_value.split(',')
                    file_path = os.path.join(root_dir, prop_value[0].strip())
                    file_offset = int(prop_value.strip(), 0) if len(prop_value) > 1 else 0
                    file_size = int(prop_value.strip(), 0) if len(prop_value) > 2 else 0
                    if file_path is None or not os.path.exists(file_path):
                        raise Exception("File path doesn't exist: {}".format(file_path))
                    with open(file_path, "rb") as f:
                        f.seek(file_offset)
                        prop_data = f.read(file_size) if file_size > 0 else f.read()
                    prop_obj = PropIncBin(prop_name, prop_data, os.path.split(file_path)[1])
                elif prop_value.startswith('/plugin/'):
                    raise NotImplementedError("Not implemented property value: /plugin/")
                elif prop_value.startswith('/bits/'):
                    raise NotImplementedError("Not implemented property value: /bits/")
                else:
                    prop_obj = PropStrings(prop_name)
                    for prop in prop_value.split('",'):
                        prop = prop.replace('"', "")
                        prop = prop.strip()
                        prop_obj.append(prop)
            if curnode is not None:
                curnode.append(prop_obj)

    return fdt_obj


def parse_dtb(data: bytes, offset: int = 0) -> FDT:
    """
    Parse FDT Binary Blob and create FDT Object
    
    :param data: FDT Binary Blob as bytes or bytearray
    :param offset:
    """
    assert isinstance(data, (bytes, bytearray)), "Invalid argument type"

    from struct import unpack_from

    fdt_obj = FDT()
    # parse header
    fdt_obj.header = Header.parse(data)
    # parse entries
    index = fdt_obj.header.off_mem_rsvmap
    while True:
        entrie = dict(zip(('address', 'size'), unpack_from(">QQ", data, offset + index)))
        index += 16
        if entrie['address'] == 0 and entrie['size'] == 0:
            break
        fdt_obj.entries.append(entrie)
    # parse nodes
    current_node = None
    fdt_obj.root = None
    index = fdt_obj.header.off_dt_struct
    while True:
        if len(data) < (offset + index + 4):
            raise Exception("Index out of range !")
        tag = unpack_from(">I", data, offset + index)[0]
        index += 4
        if tag == DTB_BEGIN_NODE:
            node_name = extract_string(data, offset + index)
            index = ((index + len(node_name) + 4) & ~3)
            if not node_name:
                node_name = '/'
            new_node = Node(node_name)
            if fdt_obj.root is None:
                fdt_obj.root = new_node
            if current_node is not None:
                current_node.append(new_node)
            current_node = new_node
        elif tag == DTB_END_NODE:
            if current_node is not None:
                current_node = current_node.parent
        elif tag == DTB_PROP:
            prop_size, prop_string_pos, = unpack_from(">II", data, offset + index)
            prop_start = index + 8
            if fdt_obj.header.version < 16 and prop_size >= 8:
                prop_start = ((prop_start + 7) & ~0x7)
            prop_name = extract_string(data, fdt_obj.header.off_dt_strings + prop_string_pos)
            prop_raw_value = data[offset + prop_start:offset + prop_start + prop_size]
            index = prop_start + prop_size
            index = ((index + 3) & ~0x3)
            if current_node is not None:
                current_node.append(new_property(prop_name, prop_raw_value))
        elif tag == DTB_END:
            break
        else:
            raise Exception("Unknown Tag: {}".format(tag))

    return fdt_obj


def diff(fdt1: FDT, fdt2: FDT) -> tuple:
    """ 
    Compare two flattened device tree objects and return list of 3 objects (same in 1 and 2, specific for 1, specific for 2)
    
    :param fdt1: The object 1 of FDT
    :param fdt2: The object 2 of FDT
    """
    assert isinstance(fdt1, FDT), "Invalid argument type"
    assert isinstance(fdt2, FDT), "Invalid argument type"

    fdt_a = FDT(fdt1.header)
    fdt_b = FDT(fdt2.header)

    if fdt1.header.version is not None and fdt2.header.version is not None:
        fdt_same = FDT(fdt1.header if fdt1.header.version > fdt2.header.version else fdt2.header)
    else:
        fdt_same = FDT(fdt1.header)

    if fdt1.entries and fdt2.entries:
        for entry_a in fdt1.entries:
            for entry_b in fdt2.entries:
                if entry_a['address'] == entry_b['address'] and entry_a['size'] == entry_b['size']:
                    fdt_same.entries.append(entry_a)
                    break

    for entry_a in fdt1.entries:
        found = False
        for entry_s in fdt_same.entries:
            if entry_a['address'] == entry_s['address'] and entry_a['size'] == entry_s['size']:
                found = True
                break
        if not found:
            fdt_a.entries.append(entry_a)

    for entry_b in fdt2.entries:
        found = False
        for entry_s in fdt_same.entries:
            if entry_b['address'] == entry_s['address'] and entry_b['size'] == entry_s['size']:
                found = True
                break
        if not found:
            fdt_b.entries.append(entry_b)

    for path, nodes, props in fdt1.walk():
        try:
            rnode = fdt2.get_node(path)
        except:
            rnode = None

        for node_b in nodes:
            if rnode is None or rnode.get_subnode(node_b.name) is None:
                fdt_a.add_item(Node(node_b.name), path)
            else:
                fdt_same.add_item(Node(node_b.name), path)

        for prop_a in props:
            if rnode is not None and prop_a == rnode.get_property(prop_a.name):
                fdt_same.add_item(prop_a.copy(), path)
            else:
                fdt_a.add_item(prop_a.copy(), path)

    for path, nodes, props in fdt2.walk():
        try:
            rnode = fdt_same.get_node(path)
        except:
            rnode = None

        for node_b in nodes:
            if rnode is None or rnode.get_subnode(node_b.name) is None:
                fdt_b.add_item(Node(node_b.name), path)

        for prop_b in props:
            if rnode is None or prop_b != rnode.get_property(prop_b.name):
                fdt_b.add_item(prop_b.copy(), path)

    return fdt_same, fdt_a, fdt_b
